import numpy as np

__all__ = ['KalmanFilter']


class KalmanFilter:
    acc_std_space = 10
    acc_std_bbox = 0.1
    meas_std_bbox = 0.05

    def __init__(self, acc_std, meas_std, z, time):
        self.dim_z = len(z)
        self.time = time
        self.acc_std = acc_std
        self.meas_std = meas_std

        # the observation matrix
        self.H = np.eye(self.dim_z, 3 * self.dim_z)

        self.x = np.vstack((z, np.zeros((2 * self.dim_z, 1))))
        self.P = np.zeros((3 * self.dim_z, 3 * self.dim_z))
        i, j = np.indices((3 * self.dim_z, 3 * self.dim_z))
        self.P[(i - j) % self.dim_z == 0] = 1e5  # initial vector is a guess -> high estimate uncertainty

    def predict(self, dt):
        # the state transition matrix -> assuming acceleration is constant
        F = np.eye(3 * self.dim_z)
        np.fill_diagonal(F[:2 * self.dim_z, self.dim_z:], dt)
        np.fill_diagonal(F[:self.dim_z, 2 * self.dim_z:], dt ** 2 / 2)

        # the process noise matrix
        A = np.zeros((3 * self.dim_z, 3 * self.dim_z))
        np.fill_diagonal(A[2 * self.dim_z:, 2 * self.dim_z:], 1)
        Q = self.acc_std ** 2 * F @ A @ F.T

        self.x = F @ self.x
        self.P = F @ self.P @ F.T + Q

    def update(self, z):
        if z is None:
            return

        # the measurement uncertainty
        R = self.meas_std ** 2 * np.eye(self.dim_z)

        # the Kalman Gain
        K = self.P @ self.H.T @ np.linalg.inv(self.H @ self.P @ self.H.T + R)

        self.x = self.x + K @ (z - self.H @ self.x)
        I = np.eye(3 * self.dim_z)
        self.P = (I - K @ self.H) @ self.P @ (I - K @ self.H).T + K @ R @ K.T



import os
from pathlib import Path
from string import Template
from typing import Tuple, Optional, List

import depthai as dai

from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.types import GenericNeuralNetwork


class MultiStageNN:
    def __init__(self,
                 pipeline: dai.Pipeline,
                 detection_node: GenericNeuralNetwork,  # Object detection node
                 high_res_frames: dai.Node.Output,
                 size: Tuple[int, int],
                 frame_size: Tuple[int, int],
                 det_nn_size: Tuple[int, int],
                 resize_mode: ResizeMode,
                 num_frames_pool: int = 20,
                 ) -> None:
        """
        Args:
            pipeline (dai.Pipeline): Pipeline object
            detection_node (GenericNeuralNetwork): Object detection NN
            high_res_frames (dai.Node.Output): Frames corresponding to the detection NN
            size (Tuple[int, int]): NN input size of the second stage NN
            frame_size (Tuple[int, int]): Frame size of the first (detection) stage NN
            det_nn_size (Tuple[int, int]): NN input size of the first (detection) stage NN
            resize_mode (ResizeMode): Resize mode that was used to resize frame for first (detection) stage
            debug (bool, optional): Enable debug mode. Defaults to False.
            num_frames_pool (int, optional): Number of frames to keep in the pool. Defaults to 20.
        """
        frame_size_ar = frame_size[0] / frame_size[1]
        det_nn_size_ar = det_nn_size[0] / det_nn_size[1]
        if resize_mode == ResizeMode.LETTERBOX:
            padding = (frame_size_ar - det_nn_size_ar) / 2
            if padding > 0:
                self.init = f"xmin = 0; ymin = {-padding}; xmax = 1; ymax = {1 + padding}"
            else:
                self.init = f"xmin = {padding}; ymin = 0; xmax = {1 - padding}; ymax = 1"
        elif resize_mode in [ResizeMode.CROP, ResizeMode.FULL_CROP]:
            cropping = (1 - (det_nn_size_ar / frame_size_ar)) / 2
            if cropping < 0:
                self.init = f"xmin = 0; ymin = {-cropping}; xmax = 1; ymax = {1 + cropping}"
            else:
                self.init = f"xmin = {cropping}; ymin = 0; xmax = {1 - cropping}; ymax = 1"
        else:  # Stretch
            self.init = f"xmin=0; ymin=0; xmax=1; ymax=1"

        self.script: dai.node.Script = pipeline.create(dai.node.Script)
        self.script.setProcessor(dai.ProcessorType.LEON_CSS)  # More stable
        self._size: Tuple[int, int] = size

        detection_node.out.link(self.script.inputs['detections'])
        high_res_frames.link(self.script.inputs['frames'])

        self.configure()  # User might later call this again with different parameters

        self.manip: dai.node.ImageManip = pipeline.create(dai.node.ImageManip)
        self.manip.initialConfig.setResize(size)
        self.manip.setWaitForConfigInput(True)
        self.manip.setMaxOutputFrameSize(size[0] * size[1] * 3)
        self.manip.setNumFramesPool(num_frames_pool)
        self.script.outputs['manip_cfg'].link(self.manip.inputConfig)
        self.script.outputs['manip_img'].link(self.manip.inputImage)
        self.out: dai.Node.Output = self.manip.out

        self.whitelist_labels: Optional[List[int]] = None
        self.scale_bb: Optional[Tuple[int, int]] = None

    def configure(self,
                  debug: bool = False,
                  whitelist_labels: Optional[List[int]] = None,
                  scale_bb: Optional[Tuple[int, int]] = None) -> None:
        """
        Args:
            debug (bool, optional): Enable debug mode. Defaults to False.
            whitelist_labels (Optional[List[int]], optional): List of labels to keep. Defaults to None.
            scale_bb (Optional[Tuple[int, int]], optional): Scale bounding box. Defaults to None.
        """
        # Used later for visualization
        self.whitelist_labels = whitelist_labels
        self.scale_bb = scale_bb

        with open(Path(os.path.dirname(__file__)) / 'template_multi_stage_script.py', 'r') as file:
            code = Template(file.read()).substitute(
                DEBUG='' if debug else '#',
                CHECK_LABELS=f"if det.label not in {str(whitelist_labels)}: continue" if whitelist_labels else "",
                WIDTH=str(self._size[0]),
                HEIGHT=str(self._size[1]),
                INIT=self.init,
                SCALE_BB_XMIN=f"-{scale_bb[0] / 100}" if scale_bb else "",  # % to float value
                SCALE_BB_YMIN=f"-{scale_bb[1] / 100}" if scale_bb else "",
                SCALE_BB_XMAX=f"+{scale_bb[0] / 100}" if scale_bb else "",
                SCALE_BB_YMAX=f"+{scale_bb[1] / 100}" if scale_bb else "",
            )
            self.script.setScript(code)
            # print(f"\n------------{code}\n---------------")



import os
from fractions import Fraction
from pathlib import Path
from typing import Tuple, Union, Optional, List

import depthai as dai
import numpy as np

from depthai_sdk.recorders.video_writers.utils import create_writer_dir
from .base_writer import BaseWriter

START_CODE_PREFIX = b"\x00\x00\x01"
KEYFRAME_NAL_TYPE = 5
NAL_TYPE_BITS = 5  # nal unit type is encoded in lower 5 bits


def is_keyframe(encoded_frame: np.array) -> bool:
    """
    Check if encoded frame is a keyframe.
    Args:
        encoded_frame: Encoded frame.

    Returns:
        True if encoded frame is a keyframe, False otherwise.
    """
    byte_stream = bytearray(encoded_frame)
    size = len(byte_stream)

    pos = 0
    while pos < size:
        retpos = byte_stream.find(START_CODE_PREFIX, pos)
        if retpos == -1:
            return False

        # Skip start code
        pos = retpos + 3

        # Extract the first 5 bits
        type_ = byte_stream[pos] >> 3

        if type_ == KEYFRAME_NAL_TYPE:
            return True

    return False


class AvWriter(BaseWriter):
    def __init__(self, path: Path, name: str, fourcc: str):
        """
        Args:
            path: Path to the folder where the file will be created.
            name: Name of the file without extension.
            fourcc: Stream codec.
            fps: Frames per second of the stream.
            frame_shape: Width and height of the frames.
        """
        super().__init__(path, name)

        self.start_ts = None
        self._fourcc = fourcc

        self._stream = None
        self._file = None
        self.closed = False
        self._codec = None  # Used to determine dimensions of encoded frames
        self._frame_buffer: List[dai.ImgFrame] = []

    def _create_stream(self, shape: Tuple) -> None:
        """
        Create stream in file with given fourcc and fps, works in-place.

        Args:
            fourcc: Stream codec.
            fps: Frames per second of the stream.
        """
        self._stream = self._file.add_stream(self._fourcc)
        self._stream.time_base = Fraction(1, 1000 * 1000)  # Microseconds

        # We need to set pixel format for MJEPG, for H264/H265 it's yuv420p by default
        if self._fourcc == 'mjpeg':
            self._stream.pix_fmt = 'yuvj420p'

        self._stream.width = shape[0]
        self._stream.height = shape[1]

    def get_dimension(self, img: dai.ImgFrame) -> Optional[Tuple[int, int]]:
        enc_packets = self._codec.parse(img.getData())
        if len(enc_packets) == 0:
            return None
        frames = self._codec.decode(enc_packets[-1])
        if not frames:
            return None
        return frames[0].width, frames[0].height

    def create_file_for_buffer(self, subfolder: str, buf_name: str) -> None:  # independent of type of frames
        self.create_file(subfolder)

    def create_file(self, subfolder: str) -> None:
        """
        Create file for writing frames.

        Args:
            subfolder: Subfolder relative to the main folder where the file will be created.
        """
        path_to_file = create_writer_dir(self.path / subfolder, self.name, 'mp4')
        self._create_file(path_to_file)

    def _create_file(self, path_to_file: str) -> None:
        """
        Create av container for writing frames.

        Args:
            path_to_file: Path to the file.
        """
        global av
        import av
        # We will remux .h264 later
        suffix = '.h264' if self._fourcc.lower() == 'h264' else '.mp4'
        self._file = av.open(str(Path(path_to_file).with_suffix(suffix)), 'w')

        # Needed to get dimensions from the frame. Only decode first frame.
        self._codec = av.CodecContext.create(self._fourcc, "r")

    def __mux_imgframe(self, frame: dai.ImgFrame) -> None:
        frame_data = frame.getData()

        if self.start_ts is None:
            # For H26x, wait for a keyframe
            if self._fourcc != 'mjpeg' and not is_keyframe(frame_data):
                return

        packet = av.Packet(frame_data)  # Create new packet with byte array

        # Set frame timestamp
        if self.start_ts is None:
            self.start_ts = frame.getTimestampDevice()

        ts = int((frame.getTimestampDevice() - self.start_ts).total_seconds() * 1e6)  # To microsec
        packet.dts = ts + 1  # +1 to avoid zero dts
        packet.pts = ts + 1
        packet.stream = self._stream
        self._file.mux_one(packet)  # Mux the Packet into container

    def write(self, frame: dai.ImgFrame) -> None:
        """
        Write packet bytes to h264 file.

        Args:
            frame: ImgFrame from depthai pipeline.
        """
        if self.closed:
            return
        if self._file is None:
            self.create_file(subfolder='')

        if self._stream is None:
            shape = self.get_dimension(frame)
            if shape is None:
                # Save frame, so we can mux it later when dimnesions are known
                self._frame_buffer.append(frame)
                return

            self._create_stream(shape)
            for buffered_frame in self._frame_buffer:
                self.__mux_imgframe(buffered_frame)

        self.__mux_imgframe(frame)

    def close(self) -> None:
        """
        Close the file and potentially remux it to mp4.
        """
        self.closed = True
        if self._file is not None:
            p = self._stream.encode(None)
            self._file.mux(p)
            self._file.close()

        # Remux the h264 stream to finalize the output file
        if self._fourcc == 'h264':
            self.remux_h264_video(str(self._file.name))

    def remux_h264_video(self, input_file: Union[Path, str]) -> None:
        """
        Remuxes h264 file to mp4.

        Args:
            input_file: path to h264 file.
        """

        mp4_file = str(Path(input_file).with_suffix('.mp4'))

        if input_file == mp4_file:
            mp4_file = str(Path(input_file).with_suffix('.remuxed.mp4'))

        with av.open(mp4_file, "w", format="mp4") as output_container, \
                av.open(input_file, "r", format=self._fourcc) as input_container:
            input_stream = input_container.streams[0]
            fps = input_stream.average_rate
            output_stream = output_container.add_stream(template=input_stream, rate=fps)

            output_stream.width = input_stream.width
            output_stream.height = input_stream.height

            frame_time = (1 / fps) * input_stream.time_base.denominator
            for i, packet in enumerate(input_container.demux(input_stream)):
                packet.dts = i * frame_time
                packet.pts = i * frame_time
                packet.stream = output_stream
                output_container.mux_one(packet)

        os.remove(input_file)

        if Path(mp4_file).suffix == '.remuxed.mp4':
            os.rename(mp4_file, input_file)



from datetime import timedelta
from typing import Dict, Union

import depthai as dai

from depthai_sdk.classes import DetectionPacket
from depthai_sdk.components import NNComponent
from depthai_sdk.trigger_action.triggers.abstract_trigger import Trigger

__all__ = ['DetectionTrigger']


class DetectionTrigger(Trigger):
    """
    Trigger that is activated when a certain number of detections of a certain label are detected.
    """

    def __init__(self,
                 input: NNComponent,
                 min_detections: Dict[str, int],
                 cooldown: Union[timedelta, int]):
        """
        Args:
            input: NNComponent that represents the NN node that will be used for detection.
            min_detections: Dictionary that maps labels to the minimum number of detections of that label that must be
                            detected for the trigger to be activated. Example: `{'person': 1, 'dog': 2}` means that
                            the trigger will be activated when at least 1 person and 2 dogs are detected.
            cooldown: Time that must pass after the trigger is activated before it can be activated again.
        """
        if not isinstance(input.node, dai.node.DetectionNetwork):
            raise ValueError('The input NNComponent must represent a node of type `dai.node.DetectionNetwork`')

        self.min_detections = min_detections
        # Convert all labels to uppercase
        self.min_detections = {label.upper(): number for label, number in min_detections.items()}

        for label, number in self.min_detections.items():
            if label not in input.get_labels():
                raise ValueError(f'Label "{label}" doesn\'t exist for network "{input.get_name()}". '
                                 f'Supported labels are:\n {input.get_labels()}')

            if number <= 0:
                raise ValueError('Numbers given in min_detections must be positive.')

        self.labels = input.get_labels()
        super().__init__(input.out.main, self.condition, cooldown)

    def condition(self, packet: DetectionPacket) -> bool:
        """
        Method that checks if the trigger should be activated.

        Args:
            packet: DetectionPacket that contains the detections.

        Returns:
            True if the trigger should be activated, False otherwise.
        """
        dets = packet.img_detections.detections

        dets_of_interest = dict.fromkeys(self.min_detections.keys(), 0)
        for det in dets:
            label_str = self.labels[det.label]
            if dets_of_interest.get(label_str) is not None:
                dets_of_interest[label_str] += 1

        return all(a >= b for a, b in zip(list(dets_of_interest.values()), list(self.min_detections.values())))



import itertools
import warnings
from collections import defaultdict
from functools import cached_property
from typing import List, Optional, Dict

import depthai as dai
import numpy as np

from depthai_sdk.classes.packets import DisparityPacket
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.oak_outputs.xout.xout_seq_sync import XoutSeqSync
from depthai_sdk.visualize.configs import StereoColor

try:
    import cv2
except ImportError:
    cv2 = None

IR_FLOOD_LIMIT = 765
IR_FLOOD_STEP = IR_FLOOD_LIMIT / 4
IR_DOT_LIMIT = 1200
IR_DOT_STEP = IR_DOT_LIMIT / 4


class XoutDisparity(XoutSeqSync, XoutFrames):
    def __init__(self,
                 device: dai.Device,
                 frames: StreamXout,
                 disp_factor: float,
                 aligned_frame: Optional[StreamXout],
                 fourcc: str = None,
                 colorize: StereoColor = None,
                 colormap: int = None,
                 wls_config: dict = None,
                 ir_settings: dict = None,
                 confidence_map: StreamXout = None):
        self.aligned_frame = aligned_frame
        self.name = 'Disparity'
        self.multiplier = disp_factor
        self.device = device

        self.confidence_map = confidence_map
        self.fig, self.axes = None, None  # for depth score visualization

        self.colorize = colorize
        self.colormap = colormap

        self.ir_settings = ir_settings
        self._dot_projector_brightness = 0  # [0, 1200]
        self._flood_brightness = 0  # [0, 1500]

        self._metrics_buffer = defaultdict(list)
        self._auto_ir_converged = False
        self._checking_neighbourhood = False
        self._converged_metric_value = None

        # Values that will be tested for function approximation
        self._candidate_pairs = list(itertools.product(np.arange(0, IR_DOT_LIMIT + 1, IR_DOT_STEP),
                                                       np.arange(0, IR_FLOOD_LIMIT + 1, IR_FLOOD_STEP)))
        self._neighbourhood_pairs = []
        self._candidate_idx, self._neighbour_idx = 0, 0
        self._X, self._y = [], []

        # Prefer to use WLS level if set, otherwise use lambda and sigma
        self.use_wls_filter = wls_config['enabled'] if wls_config else False
        if self.use_wls_filter:
            wls_level = wls_config['level']
            LOGGER.debug(
                f'Using WLS level: {wls_level.name} (lambda: {wls_level.value[0]}, sigma: {wls_level.value[1]})'
            )
            self.wls_lambda = wls_level.value[0] or wls_config['lambda']
            self.wls_sigma = wls_level.value[1] or wls_config['sigma']

            try:
                self.wls_filter = cv2.ximgproc.createDisparityWLSFilterGeneric(False)
                self.wls_filter.setLambda(self.wls_lambda)
                self.wls_filter.setSigmaColor(self.wls_sigma)

            except AttributeError:
                warnings.warn(
                    'OpenCV version does not support WLS filter. Disabling WLS filter. '
                    'Make sure you have opencv-contrib-python installed. '
                    'If not, run "pip uninstall opencv-python && pip install opencv-contrib-python -U"'
                )
                self.use_wls_filter = False

        XoutFrames.__init__(self, frames=frames, fourcc=fourcc)
        XoutSeqSync.__init__(self, self.xstreams())

    def on_callback(self, packet) -> None:
        if self.ir_settings['auto_mode']:
            self._auto_ir_search(packet.msg.getFrame())

    def xstreams(self) -> List[StreamXout]:
        streams = [self.frames]
        if self.aligned_frame is not None:
            streams.append(self.aligned_frame)
        if self.confidence_map is not None:
            streams.append(self.confidence_map)

        return streams

    def package(self, msgs: Dict) -> DisparityPacket:
        img_frame = msgs[self.frames.name]
        aligned_frame = msgs[self.aligned_frame.name] if self.aligned_frame else None
        confidence_map = msgs[self.confidence_map.name] if self.confidence_map else None
        # TODO: refactor the mess below
        packet = DisparityPacket(
            self.get_packet_name(),
            img_frame,
            self.multiplier,
            disparity_map=None,
            colorize=self.colorize,
            colormap=self.colormap,
            confidence_map=confidence_map,
            aligned_frame=aligned_frame,
        )
        packet._get_codec = self.get_codec

        if self._fourcc is None:
            disparity_frame = img_frame.getFrame()
        else:
            disparity_frame = packet.decode()
            if disparity_frame is None:
                return None

        if aligned_frame and self.use_wls_filter:
            # Perform WLS filtering
            # If we have wls enabled, it means CV2 is installed
            disparity_frame = self.wls_filter.filter(disparity_frame, aligned_frame.getCvFrame())

        packet.disparity_map = disparity_frame

        return packet

    def _auto_ir_search(self, frame: np.ndarray):
        # Perform neighbourhood search if we got worse metric values
        if self._checking_neighbourhood:
            # Increment the neighbourhood index if we have finished checking the current neighbour pair
            if self._ir_grid_search_iteration(frame, self._neighbourhood_pairs, self._neighbour_idx):
                self._neighbour_idx += 1

        # Check if we have finished checking all candidates, done once on the start up
        elif not self._auto_ir_converged:
            # Increment the candidate index if we have finished checking the current candidate pair
            if self._ir_grid_search_iteration(frame, self._candidate_pairs, self._candidate_idx):
                self._candidate_idx += 1

        # Continuously check the consistency of the metric values, if we are in continuous mode
        elif self._auto_ir_converged and self.ir_settings['continuous_mode']:
            self._check_consistency(frame)

    def _ir_grid_search_iteration(self, frame: np.array, candidate_pairs: list = None, candidate_idx: int = 0):
        fill_rate = np.count_nonzero(frame) / frame.size
        self._metrics_buffer['fill_rate'].append(fill_rate)

        if len(self._metrics_buffer['fill_rate']) < 30:
            return False

        if candidate_idx >= len(candidate_pairs):
            # We have exhausted all candidates
            best_idx = np.argmax(self._y)
            self._converged_metric_value = self._y[best_idx]
            self._dot_projector_brightness, self._flood_brightness = self._X[best_idx]
            self._reset_buffers()
            self._auto_ir_converged = True
            self._checking_neighbourhood = False

            LOGGER.debug(f'Auto IR converged: dot projector - {self._dot_projector_brightness}mA, '
                          f'flood - {self._flood_brightness}mA')
        else:
            self._dot_projector_brightness, self._flood_brightness = candidate_pairs[candidate_idx]

        self._update_ir()

        if self._auto_ir_converged:
            return False

        # Skip first half second of frames to allow for auto exposure to settle down
        fill_rate_avg = np.mean(self._metrics_buffer['fill_rate'][15:])

        self._X.append([self._dot_projector_brightness, self._flood_brightness])
        self._y.append(fill_rate_avg)

        self._metrics_buffer['fill_rate'].clear()
        return True

    def _check_consistency(self, frame):
        fill_rate = np.count_nonzero(frame) / frame.size
        self._metrics_buffer['fill_rate'].append(fill_rate)

        if len(self._metrics_buffer['fill_rate']) < 30:
            return

        fill_rate_avg = np.mean(self._metrics_buffer['fill_rate'])
        self._metrics_buffer['fill_rate'].clear()

        if fill_rate_avg < self._converged_metric_value * 0.85:
            self._auto_ir_converged = False
            self._checking_neighbourhood = True
            self._neighbourhood_pairs = np.unique([
                [np.clip(self._dot_projector_brightness + i, 0, IR_DOT_LIMIT),
                 np.clip(self._flood_brightness + j, 0, IR_FLOOD_LIMIT)]
                for i, j in itertools.product([-IR_DOT_STEP / 2, IR_DOT_STEP / 2],
                                              [-IR_FLOOD_STEP / 2, IR_FLOOD_STEP / 2])
            ], axis=0)
            self._neighbour_idx = 0

    def _update_ir(self):
        self.device.setIrLaserDotProjectorBrightness(self._dot_projector_brightness)
        self.device.setIrFloodLightBrightness(self._flood_brightness)

    def _reset_buffers(self):
        self._X, self._y = [], []
        del self._metrics_buffer['fill_rate']



from typing import List, Optional

from depthai_sdk.classes.packets import FramePacket
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout

try:
    import cv2
except ImportError:
    cv2 = None


class XoutFrames(XoutBase):
    """
    Stream of frames. Single message, no syncing required.
    """

    def __init__(self,
                 frames: StreamXout,
                 fourcc: Optional[str] = None,  # 'mjpeg', 'h264', 'hevc'
                 ):
        """
        Args:
            frames: StreamXout object.
            fourcc: Codec to use for encoding. If None, no encoding will be done.
        """
        self.frames = frames
        self.name = frames.name
        self._fourcc = fourcc
        self._codec = None

        super().__init__()

    def set_fourcc(self, fourcc: str) -> 'XoutFrames':
        self._fourcc = fourcc
        return self

    def xstreams(self) -> List[StreamXout]:
        return [self.frames]

    def new_msg(self, name: str, msg):
        if name not in self._streams:
            return
        return FramePacket(self.get_packet_name(), msg)

    def get_codec(self):
        # No codec, frames are NV12/YUV/BGR, so we can just use imgFrame.getCvFrame()
        if self._fourcc is None:
            return None

        if self._codec is None:
            try:
                import av
            except ImportError:
                raise ImportError('Attempted to decode an encoded frame, but av is not installed.'
                                  ' Please install it with `pip install av`')
            self._codec = av.CodecContext.create(self._fourcc, "r")
        return self._codec



from depthai_sdk.args_parser import ArgsParser
from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.constants import CV2_HAS_GUI_SUPPORT
from depthai_sdk.logger import set_logging_level
from depthai_sdk.oak_camera import OakCamera
from depthai_sdk.previews import *
from depthai_sdk.record import *
from depthai_sdk.replay import *
from depthai_sdk.utils import *
from depthai_sdk.utils import _create_config, get_config_field, _sentry_before_send
from depthai_sdk.visualize import *

__version__ = '1.15'


def __import_sentry(sentry_dsn: str) -> None:
    try:
        import sentry_sdk

        sentry_sdk.init(
            dsn=sentry_dsn,
            traces_sample_rate=1.0,
            release=f'depthai_sdk@{__version__}',
            with_locals=False,
            before_send=_sentry_before_send
        )
    except:
        pass


sentry_dsn = get_config_field('sentry_dsn')
sentry_status = get_config_field('sentry')
if sentry_dsn and sentry_status:
    __import_sentry(sentry_dsn)



#!/usr/bin/env python3
from enum import IntEnum
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import List

import depthai as dai

from depthai_sdk.classes.packets import FramePacket, IMUPacket
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.recorders.abstract_recorder import Recorder

def _run(recorder: Recorder, frame_queue: Queue):
    """
    Start recording infinite loop
    """
    while True:
        try:
            frames = frame_queue.get()
            if frames is None:  # Terminate app
                break

            for name in frames:
                # Save all synced frames into files
                recorder.write(name, frames[name])
        except KeyboardInterrupt:
            break
    # Close all recorders - Can't use ExitStack with VideoWriter
    recorder.close()
    LOGGER.info('Exiting store frame thread')


class RecordType(IntEnum):
    VIDEO = 1  # Save to video file
    VIDEO_LOSSLESS = 2  # Save to lossless video file (.avi)
    ROSBAG = 3  # To ROS .bag
    MCAP = 4  # To .mcap
    DB3 = 5  # To .db3 (ros2)


class Record:
    """
    This class records depthai streams from OAK cameras into different formats.
    It will also save calibration .json, so depth reconstruction will be possible.
    """

    def __init__(self, path: Path, record_type: RecordType):
        """
        Args:
            path (Path): Path to the recording folder
            record_type (RecordType): Recording type
        """
        self.folder = path
        self.record_type = record_type
        self.frame_q = None
        self.name_mapping = None  # XLinkOut stream name -> Friendly name mapping

        self.stream_num = None
        self.mxid = None
        self.path = None
        self.process = None

        if self.record_type == RecordType.MCAP:
            from .recorders.mcap_recorder import McapRecorder
            self.recorder = McapRecorder()
        elif self.record_type == RecordType.VIDEO:
            from .recorders.video_recorder import VideoRecorder
            self.recorder = VideoRecorder()
        elif self.record_type == RecordType.VIDEO_LOSSLESS:
            from .recorders.video_recorder import VideoRecorder
            self.recorder = VideoRecorder(lossless=True)
        elif self.record_type == RecordType.ROSBAG:
            from .recorders.rosbag_recorder import Rosbag1Recorder
            self.recorder = Rosbag1Recorder()
        elif self.record_type == RecordType.DB3:
            from .recorders.rosbag_recorder import Rosbag2Recorder
            self.recorder = Rosbag2Recorder()
        else:
            raise ValueError(f"Recording type '{self.record_type}' isn't supported!")

    def write(self, packets):
        if not isinstance(packets, dict):
            packets = {packets.name: packets}

        msgs = dict()
        for name, packet in packets.items():
            if isinstance(packet, FramePacket):
                msgs[name] = packet.msg
            elif isinstance(packet, IMUPacket):
                msgs[name] = packet.packet
        self.frame_q.put(msgs)

    def start(self, device: dai.Device, xouts: List[XoutFrames]):
        """
        Start recording process. This will create and start the pipeline,
        start recording threads, and initialize all queues.
        """
        self.mxid = device.getMxId()
        self.path = self._create_folder(self.folder, self.mxid)
        calib_data = device.readCalibration()
        calib_data.eepromToJsonFile(str(self.path / "calib.json"))

        self.recorder.update(self.path, device, xouts)

        self.frame_q = Queue(maxsize=20)
        self.process = Thread(target=_run, args=(self.recorder, self.frame_q))
        self.process.start()

    # TODO: support pointclouds in MCAP
    def config_mcap(self, pointcloud: bool):
        if self.record_type != RecordType.MCAP:
            LOGGER.info(f"Recorder type is {self.record_type}, not MCAP! Config attempt ignored.")
            return
        self.recorder.set_pointcloud(pointcloud)

    # TODO: implement config of BAG to either record depth as frame or pointcloud
    # def config_bag(self, pointcloud: bool):
    #     if self.type != RecordType.BAG:
    #         print(f"Recorder type is {self.type}, not BAG! Config attempt ignored.")
    #     self.recorder.set_pointcloud(pointcloud)

    def _create_folder(self, path: Path, mxid: str) -> Path:
        """
        Creates recording folder
        """
        i = 0
        while True:
            i += 1
            recordings_path = path / f"{i}-{str(mxid)}"
            if not recordings_path.is_dir():
                recordings_path.mkdir(parents=True, exist_ok=False)
                return recordings_path

    def close(self):
        if self.frame_q:
            self.frame_q.put(None)  # Close recorder and stop the thread



#
# from typing import Tuple
#
# import depthai as dai
# import numpy as np
# import std_msgs
# from genpy.rostime import Time
# # from std_msgs.msg import Header, ColorRGBA, String
# # from visualization_msgs.msg import ImageMarker
# from geometry_msgs.msg import TransformStamped
# from geometry_msgs.msg import Vector3, Quaternion, Transform
# from sensor_msgs.msg import CompressedImage, Image, PointCloud2, PointField, Imu  # s, PointCloud
# from tf.msg import tfMessage
# from .imu_interpolation import ImuSyncMethod
#
# """
# --extra-index-url https://rospypi.github.io/simple/
# sensor_msgs
# geometry_msgs
# std_msgs
# genpy
# tf2_msgs
# tf2_ros
# tf
# """
#
#
# class DepthAi2Ros1:
#     def __init__(self, device: dai.Device) -> None:
#         self.start_time = dai.Clock.now()
#         self.device = device
#
#         self.xyz = dict()
#
#     def header(self, msg: dai.Buffer) -> std_msgs.msg.Header:
#         header = std_msgs.msg.Header()
#         ts = (msg.getTimestamp() - self.start_time).total_seconds()
#         # secs / nanosecs
#         header.stamp = Time(int(ts), (ts % 1) * 1e6)
#         try:
#             header.frame_id = str(msg.getSequenceNum())  # ImgFrame
#         except:
#             header.frame_id = str(msg.sequence)  # IMUReport
#         return header
#
#     def CompressedImage(self, imgFrame: dai.ImgFrame) -> CompressedImage:
#         msg = CompressedImage()
#         msg.header = self.header(imgFrame)
#         msg.format = "jpeg"
#         msg.data = imgFrame.getData().tobytes()
#         return msg
#
#     def Imu(self, imu_packet: dai.IMUPacket, sync_mode: ImuSyncMethod = ImuSyncMethod.LINEAR_INTERPOLATE_ACCEL,
#             linear_accel_cov: float = 0., angular_velocity_cov: float = 0) -> Imu:
#         if len(self.imu_packets) > 20:
#             self.imu_packets.pop(0)
#
#         self.imu_packets.append(imu_packet)
#
#         if sync_mode != ImuSyncMethod.COPY:
#             interp_imu_packets = self.fillImuData_LinearInterpolation(self, sync_mode)
#             if len(interp_imu_packets) > 0:
#                 imu_packet = interp_imu_packets[-1]
#
#         msg = Imu()
#
#         if imu_packet.acceleroMeter is not None:
#             msg.header = self.header(imu_packet.acceleroMeter)
#             msg.linear_acceleration.x = imu_packet.acceleroMeter.x
#             msg.linear_acceleration.y = imu_packet.acceleroMeter.y
#             msg.linear_acceleration.z = imu_packet.acceleroMeter.z
#
#         if imu_packet.gyroscope is not None:
#             msg.header = self.header(imu_packet.gyroscope)
#             msg.angular_velocity.x = imu_packet.gyroscope.x
#             msg.angular_velocity.y = imu_packet.gyroscope.y
#             msg.angular_velocity.z = imu_packet.gyroscope.z
#
#         msg.orientation.x = 0.0
#         msg.orientation.y = 0.0
#         msg.orientation.z = 0.0
#         msg.orientation.w = 1.0
#
#         msg.orientation_covariance = [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
#         msg.linear_acceleration_covariance = [linear_accel_cov, 0.0, 0.0, 0.0, linear_accel_cov, 0.0, 0.0, 0.0,
#                                               linear_accel_cov]
#         msg.angular_velocity_covariance = [angular_velocity_cov, 0.0, 0.0, 0.0, angular_velocity_cov, 0.0, 0.0, 0.0,
#                                            angular_velocity_cov]
#
#         return msg
#
#
#
#     def Image(self, imgFrame: dai.ImgFrame) -> Image:
#         msg = Image()
#         msg.header = self.header(imgFrame)
#         msg.height = imgFrame.getHeight()
#         msg.width = imgFrame.getWidth()
#         msg.step = imgFrame.getWidth()
#         msg.is_bigendian = 0
#
#         type = imgFrame.getType()
#         TYPE = dai.ImgFrame.Type
#         if type == TYPE.RAW16:  # Depth
#             msg.encoding = 'mono16'
#             msg.step *= 2  # 2 bytes per pixel
#             msg.data = imgFrame.getData().tobytes()
#         elif type in [TYPE.GRAY8, TYPE.RAW8]:  # Mono frame
#             msg.encoding = 'mono8'
#             msg.data = imgFrame.getData().tobytes()  # np.array(imgFrame.getFrame()).tobytes()
#         else:
#             msg.encoding = 'bgr8'
#             msg.data = imgFrame.getCvFrame().tobytes()
#         return msg
#
#     def TfMessage(self,
#                   imgFrame: dai.ImgFrame,
#                   translation: Tuple[float, float, float] = (0., 0., 0.),
#                   rotation: Tuple[float, float, float, float] = (0., 0., 0., 0.)) -> tfMessage:
#         msg = tfMessage()
#         tf = TransformStamped()
#         tf.header = self.header(imgFrame)
#         tf.child_frame_id = str(imgFrame.getSequenceNum())
#         tf.transform = Transform(
#             translation=Vector3(x=translation[0], y=translation[1], z=translation[2]),
#             rotation=Quaternion(x=rotation[0], y=rotation[1], z=rotation[2], w=rotation[3])
#         )
#         msg.transforms.append(tf)
#         return msg
#
#     def PointCloud2(self, imgFrame: dai.ImgFrame) -> PointCloud2:
#         """
#         Depth frame -> ROS1 PointCloud2 message
#         """
#         msg = PointCloud2()
#         msg.header = self.header(imgFrame)
#
#         heigth = str(imgFrame.getHeight())
#         if heigth not in self.xyz:
#             self._create_xyz(imgFrame.getWidth(), imgFrame.getHeight())
#
#         frame = imgFrame.getCvFrame()
#         frame = np.expand_dims(np.array(frame), axis=-1)
#         pcl = self.xyz[heigth] * frame / 1000.0  # To meters
#
#         msg.height = imgFrame.getHeight()
#         msg.width = imgFrame.getWidth()
#         msg.fields = [
#             PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
#             PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
#             PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1)
#         ]
#         msg.is_bigendian = False
#         msg.point_step = 12  # 3 * float32 (=4 bytes)
#         msg.row_step = 12 * imgFrame.getWidth()
#         msg.data = pcl.tobytes()
#         msg.is_dense = True
#         return msg
#
#     def _create_xyz(self, width, height):
#         calibData = self.device.readCalibration()
#         M_right = calibData.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(width, height))
#         camera_matrix = np.array(M_right).reshape(3, 3)
#
#         xs = np.linspace(0, width - 1, width, dtype=np.float32)
#         ys = np.linspace(0, height - 1, height, dtype=np.float32)
#
#         # generate grid by stacking coordinates
#         base_grid = np.stack(np.meshgrid(xs, ys))  # WxHx2
#         points_2d = base_grid.transpose(1, 2, 0)  # 1xHxWx2
#
#         # unpack coordinates
#         u_coord: np.array = points_2d[..., 0]
#         v_coord: np.array = points_2d[..., 1]
#
#         # unpack intrinsics
#         fx: np.array = camera_matrix[0, 0]
#         fy: np.array = camera_matrix[1, 1]
#         cx: np.array = camera_matrix[0, 2]
#         cy: np.array = camera_matrix[1, 2]
#
#         # projective
#         x_coord: np.array = (u_coord - cx) / fx
#         y_coord: np.array = (v_coord - cy) / fy
#
#         xyz = np.stack([x_coord, y_coord], axis=-1)
#         self.xyz[str(height)] = np.pad(xyz, ((0, 0), (0, 0), (0, 1)), "constant", constant_values=1.0)



${INIT}
width=xmax-xmin; height=ymax-ymin;
while True:
    detections = node.io['detections'].get()
    if len(detections.detections) == 0:
        continue
    largest_det = detections.detections[0]
    largest_size = 0
    for det in detections.detections:
        size = (det.xmax - det.xmin) * (det.ymax - det.ymin)
        if size > largest_size:
            largest_size = size
            largest_det = det
    det = largest_det
    ${DEBUG}node.warn(f"Detected ({det.xmin}, {det.ymin}) ({det.xmax}, {det.ymax})")
    ${RESIZE}
    if new_xmin < 0: new_xmin = 0.001
    if new_ymin < 0: new_ymin = 0.001
    if new_xmax > 1: new_xmax = 0.999
    if new_ymax > 1: new_ymax = 0.999

    ${DEBUG}node.warn(f"New ({new_xmin}, {new_ymin}) ({new_xmax}, {new_ymax})")
    ${DENORMALIZE}
    ${DEBUG}node.warn(f"Denormalized START ({startx}, {starty}) Width: {new_width}, height: {new_height})")
    control = CameraControl(1)
    ${CONTROL}
    node.io['control'].send(control)


from typing import Union, Dict

import numpy as np

from .abstract_recorder import *
from depthai_sdk.logger import LOGGER


class VideoRecorder(Recorder):
    """
    Writes video streams (.mjpeg/.h264/.hevc) or directly to mp4/avi container.
    """

    def __init__(self, lossless: bool = False):
        self.path = None
        self._stream_type = dict()
        self._writers = dict()
        self._closed = False
        self._lossless = lossless

    def __getitem__(self, item):
        return self._writers[item]

    # TODO device is not used
    def update(self, path: Path, device: dai.Device, xouts: List['XoutFrames']):
        """
        Update the recorder with new streams.
        Args:
            path: Path to save the output. Either a folder or a file.
            device: Device to get the streams from.
            xouts: List of output streams.
        """
        if path is None:
            return

        self.path = path
        if path.suffix == '' and path != Path('.'):  # If no extension, create a folder
            self.path.mkdir(parents=True, exist_ok=True)

        for xout in xouts:
            # for example, 'color_bitstream' (encoded) or 'color_video' (unencoded),
            # if component was created with name='color'
            xout_name = xout.name  # for example, 'color' --> file is color.mp4 (encoded) or color.avi (unencoded)
            file_name = xout_name
            if file_name.startswith('CameraBoardSocket.'):
                file_name = file_name[len('CameraBoardSocket.'):]
            stream = OakStream(xout)
            fourcc = stream.fourcc()  # TODO add default fourcc? stream.fourcc() can be None.

            print(fourcc, xout_name, stream.type)
            if stream.is_raw() or stream.is_depth():
                from .video_writers.video_writer import VideoWriter
                self._writers[xout_name] = VideoWriter(self.path, file_name, self._lossless)
            else:
                try:
                    from .video_writers.av_writer import AvWriter
                    self._writers[xout_name] = AvWriter(self.path, file_name, fourcc)
                except Exception as e:
                    # TODO here can be other errors, not only import error
                    LOGGER.warning(f'Exception while creating AvWriter: {e}.'
                                    '\nFalling back to FileWriter, saving uncontainerized encoded streams.')
                    from .video_writers.file_writer import FileWriter
                    self._writers[xout_name] = FileWriter(self.path, file_name, fourcc)

    def create_files_for_buffer(self, subfolder: str, buf_name: str):
        for _, writer in self._writers.items():
            writer.create_file_for_buffer(subfolder, buf_name)

    def create_file_for_buffer(self, wr_name: str, subfolder: str, buf_name: str):
        self._writers[wr_name].create_file_for_buffer(subfolder, buf_name)

    def create_file(self, wr_name: str, subfolder: str, frame: Union[np.ndarray, dai.ImgFrame]):
        self._writers[wr_name].create_file(subfolder, frame)

    def init_buffers(self, buffers: Dict[str, int]):
        for _, writer in self._writers.items():
            for name, max_seconds in buffers.items():
                writer.init_buffer(name, max_seconds)

    def add_to_buffers(self, buf_name: str, frames: Dict[str, Union[np.ndarray, dai.ImgFrame]]):
        for name, writer in self._writers.items():
            writer.add_to_buffer(buf_name, frames[name])

    def add_to_buffer(self, wr_name: str, buf_name: str, frame: Union[np.ndarray, dai.ImgFrame]):
        self._writers[wr_name].add_to_buffer(buf_name, frame)

    def is_buffer_full(self, wr_name: str, buf_name: str):
        return self._writers[wr_name].is_buffer_full(buf_name)

    def is_buffer_empty(self, wr_name: str, buf_name: str):
        return self._writers[wr_name].is_buffer_empty(buf_name)

    def write_from_buffer(self, wr_name: str, buf_name: str, n_elems: int):
        self._writers[wr_name].write_from_buffer(buf_name, n_elems)

    def write(self, name: str, frame: Union[np.ndarray, dai.ImgFrame]):
        self._writers[name].write(frame)

    def close_files(self):
        for _, writer in self._writers.items():
            writer.close()

    def close(self):
        if self._closed:
            return
        self._closed = True
        LOGGER.info(f'Video Recorder saved stream(s) to folder: {str(self.path)}')
        # Close opened files
        for name, writer in self._writers.items():
            writer.close()



from typing import Union

import depthai as dai

from depthai_sdk.classes.packets import SemanticSegmentationPacket, ImgLandmarksPacket, NnOutputPacket, DetectionPacket

GenericNeuralNetwork = Union[
    dai.node.NeuralNetwork,
    dai.node.MobileNetDetectionNetwork,
    dai.node.MobileNetSpatialDetectionNetwork,
    dai.node.YoloDetectionNetwork,
    dai.node.YoloSpatialDetectionNetwork
]

XoutNNOutputPacket = Union[
    NnOutputPacket,
    DetectionPacket,
    ImgLandmarksPacket,
    SemanticSegmentationPacket
]

Resolution = Union[
    str,
    dai.ColorCameraProperties.SensorResolution,
    dai.MonoCameraProperties.SensorResolution
]

NNNode = Union[
    dai.node.NeuralNetwork,
    dai.node.MobileNetDetectionNetwork,
    dai.node.MobileNetSpatialDetectionNetwork,
    dai.node.YoloDetectionNetwork,
    dai.node.YoloSpatialDetectionNetwork
]



from itertools import cycle

import depthai as dai
from depthai_sdk.logger import LOGGER

LIMITS = {
    'exposure_compensation': (-9, 9),
    'exposure': (1, 33000),
    'gain': (100, 1600),
    'focus': (0, 255),
    'white_balance': (1000, 12000),
    'brightness': (-10, 10),
    'contrast': (-10, 10),
    'saturation': (-10, 10),
    'sharpness': (0, 4),
    'luma_denoise': (0, 4),
    'chroma_denoise': (0, 4)
}


def clamp(value, min_value, max_value):
    return max(min(value, max_value), min_value)


class CameraControl:
    def __init__(self):
        self.queue = None

        self._cycle_awb_mode = cycle(
            [item for name, item in vars(dai.CameraControl.AutoWhiteBalanceMode).items() if name.isupper()])
        self._cycle_ab_mode = cycle(
            [item for name, item in vars(dai.CameraControl.AntiBandingMode).items() if name.isupper()])
        # self._cycle_effect_mode = cycle([item for name, item in vars(dai.CameraControl.EffectMode).items() if name.isupper()])
        self._cycle_af_mode = cycle(
            [item for name, item in vars(dai.CameraControl.AutoFocusMode).items() if name.isupper()])

        self._current_vals = {
            'exposure_time': 20000,
            'sensitivity': 300,
            'focus': 120,
            'sharpness': 0,
            'luma_denoise': 0,
            'chroma_denoise': 4,
        }

    def set_input_queue(self, queue: dai.DataInputQueue):
        self.queue = queue

    def switch_white_balance_mode(self):
        """
        Switch between auto white balance modes.
        """
        mode = next(self._cycle_awb_mode)
        LOGGER.info(f'Setting white balance mode to {mode}.')
        self.send_controls({'white-balance': {'mode': mode}})

    def switch_anti_banding_mode(self):
        """
        Switch between anti-banding modes.
        """
        mode = next(self._cycle_ab_mode)
        LOGGER.info(f'Setting anti-banding mode to {mode}.')
        self.send_controls({'ab_mode': mode})

    def switch_focus_mode(self):
        """
        Switch between focus modes.
        """
        mode = next(self._cycle_af_mode)
        LOGGER.info(f'Setting focus mode to {mode}.')
        self.send_controls({'focus': {'mode': mode}})

    def exposure_time_up(self, step=500):
        """
        Increase exposure time by step.
        Args:
            step: In microseconds
        """
        if LIMITS['exposure'][1] < self._current_vals['exposure_time'] + step:
            LOGGER.error(f'Exposure time cannot be greater than {LIMITS["exposure"][1]}')
            return
        self._current_vals['exposure_time'] += step
        self.send_controls(
            {'exposure': {'manual': [self._current_vals['exposure_time'], self._current_vals['sensitivity']]}})

    def exposure_time_down(self, step=500):
        """
        Decrease exposure time by step.
        Args:
            step: In microseconds
        """
        if LIMITS['exposure'][0] > self._current_vals['exposure_time'] - step:
            LOGGER.error(f'Exposure time cannot be less than {LIMITS["exposure"][0]}')
            return
        self._current_vals['exposure_time'] -= step
        self.send_controls(
            {'exposure': {'manual': [self._current_vals['exposure_time'], self._current_vals['sensitivity']]}})

    def sensitivity_up(self, step=50):
        """
        Increase sensitivity by step.
        Args:
            step: In ISO
        """
        if LIMITS['gain'][1] < self._current_vals['sensitivity'] + step:
            LOGGER.error(f'Sensitivity cannot be greater than {LIMITS["gain"][1]}')
            return
        self._current_vals['sensitivity'] += step
        self.send_controls(
            {'exposure': {'manual': [self._current_vals['exposure_time'], self._current_vals['sensitivity']]}})

    def sensitivity_down(self, step=50):
        """
        Decrease sensitivity by step.
        Args:
            step: In ISO
        """
        if LIMITS['gain'][0] > self._current_vals['sensitivity'] - step:
            LOGGER.error(f'Sensitivity cannot be less than {LIMITS["gain"][0]}')
            return
        self._current_vals['sensitivity'] -= step
        self.send_controls(
            {'exposure': {'manual': [self._current_vals['exposure_time'], self._current_vals['sensitivity']]}})

    def focus_up(self, step=3):
        """
        Increase focus by step.
        """
        if LIMITS['focus'][1] < self._current_vals['focus'] + step:
            LOGGER.error(f'Focus cannot be greater than {LIMITS["focus"][1]}')
            return
        self._current_vals['focus'] += step
        self.send_controls({'focus': {'manual': self._current_vals['focus']}})

    def focus_down(self, step=3):
        """
        Decrease focus by step.
        """
        if LIMITS['focus'][0] > self._current_vals['focus'] - step:
            LOGGER.error(f'Focus cannot be less than {LIMITS["focus"][0]}')
            return
        self._current_vals['focus'] -= step
        self.send_controls({'focus': {'manual': self._current_vals['focus']}})

    def sharpness_up(self, step=1):
        """
        Increase sharpness by step
        """
        if LIMITS['sharpness'][1] < self._current_vals['sharpness'] + step:
            LOGGER.error(f'Sharpness cannot be greater than {LIMITS["sharpness"][1]}')
            return
        self._current_vals['sharpness'] += step
        self.send_controls({'isp': {'sharpness': self._current_vals['sharpness']}})

    def sharpness_down(self, step=1):
        """
        Decrease sharpness by step
        """
        if LIMITS['sharpness'][0] > self._current_vals['sharpness'] - step:
            LOGGER.error(f'Sharpness cannot be less than {LIMITS["sharpness"][0]}')
            return
        self._current_vals['sharpness'] -= step
        self.send_controls({'isp': {'sharpness': self._current_vals['sharpness']}})

    def luma_denoise_up(self, step=1):
        """
        Increase luma denoise by step
        """
        if LIMITS['luma_denoise'][1] < self._current_vals['luma_denoise'] + step:
            LOGGER.error(f'Luma denoise cannot be greater than {LIMITS["luma_denoise"][1]}')
            return
        self._current_vals['luma_denoise'] += step
        self.send_controls({'isp': {'luma_denoise': self._current_vals['luma_denoise']}})

    def luma_denoise_down(self, step=1):
        """
        Decrease luma denoise by step
        """
        if LIMITS['luma_denoise'][0] > self._current_vals['luma_denoise'] - step:
            LOGGER.error(f'Luma denoise cannot be less than {LIMITS["luma_denoise"][0]}')
            return
        self._current_vals['luma_denoise'] -= step
        self.send_controls({'isp': {'luma_denoise': self._current_vals['luma_denoise']}})

    def chroma_denoise_up(self, step=1):
        """
        Increase chroma denoise by step
        """
        if LIMITS['chroma_denoise'][1] < self._current_vals['chroma_denoise'] + step:
            LOGGER.error(f'Chroma denoise cannot be greater than {LIMITS["chroma_denoise"][1]}')
            return
        self._current_vals['chroma_denoise'] += step
        self.send_controls({'isp': {'chroma_denoise': self._current_vals['chroma_denoise']}})

    def chroma_denoise_down(self, step=1):
        """
        Decrease chroma denoise by step
        """
        if LIMITS['chroma_denoise'][0] > self._current_vals['chroma_denoise'] - step:
            LOGGER.error(f'Chroma denoise cannot be less than {LIMITS["chroma_denoise"][0]}')
            return
        self._current_vals['chroma_denoise'] -= step
        self.send_controls({'isp': {'chroma_denoise': self._current_vals['chroma_denoise']}})

    def send_controls(self, controls: dict = None):
        """
        Send controls to the camera. Dict structure and available options:

        {
            'exposure':{
                'auto': False, # setAutoExposureEnable
                'lock': False, # setAutoExposureLock
                'region': [0, 0.5, 0.3, 0.75], # setAutoExposureRegion
                'compensation': 0, # setExposureCompensation
                'manual': [1000, 100] # setManualExposure(exposureTimeUs, sensitivityIso [100..1600])
            },
            'ab_mode': 'auto', # either 'off', '50hz', '60hz', or 'auto'
            'focus': {
                'range': [0,255], # setAutoFocusLensRange
                'region': [0, 0.5, 0.3, 0.75], # setAutoFocusRegion
                'trigger': False, # setAutoFocusTrigger
                'mode': 'auto', # setAutoFocusMode, either 'auto', 'macro', 'continuous_video', 'continuous_picture', or 'edof'
                'manual': 0, # setManualFocus, 0..255
            },
            'white-balance': {
                'mode': 'auto', # setAutoWhiteBalanceMode, either 'off', 'auto', 'incandescent', 'fluorescent', 'warm-fluorescent', 'daylight', 'cloudy-daylight', 'twilight', or 'shade'
                'lock': False,
                'manual': 0, # setManualWhiteBalance, range 1000..12000 [colorTemperatureK]
            },
            'isp': {
                'brightness': 0, # setBrightness(), -10..10
                'contrast': 0, # setContrast(), -10..10
                'saturation': 0, # setSaturation(), -10..10
                'sharpness': 0, # setSharpness(), 0..4
                'luma_denoise': 0, # setLumaDenoise(), 0..4
                'chroma_denoise': 0, # setChromaDenoise(), 0..4
            },
            'still': True, # Capture still photo
            'reset': True, # Reset all
        }
        """
        if self.queue is None:
            LOGGER.error('Cannot send controls when replaying.')
            return

        ctrl = dai.CameraControl()

        if controls.get('reset', None) and controls['reset']:
            LOGGER.info('Resetting camera controls.')
            ctrl.setAntiBandingMode(dai.CameraControl.AntiBandingMode.AUTO)
            ctrl.setAutoExposureEnable(True)
            ctrl.setAutoFocusMode(dai.CameraControl.AutoFocusMode.CONTINUOUS_VIDEO)
            ctrl.setAutoWhiteBalanceMode(dai.CameraControl.AutoWhiteBalanceMode.AUTO)
            self.queue.send(ctrl)
            return

        if controls.get('exposure', None) is not None:
            if controls['exposure'].get('auto', False) and controls['exposure']['auto']:
                LOGGER.info(f'Setting auto exposure to {controls["exposure"]["auto"]}.')
                ctrl.setAutoExposureEnable()
            if controls['exposure'].get('lock', None) is not None:
                LOGGER.info(f'Setting exposure lock to {controls["exposure"]["lock"]}.')
                ctrl.setAutoExposureLock(controls['exposure']['lock'])
            if controls['exposure'].get('region', None) is not None:
                # LOGGER.info(f'Setting exposure region to {controls["exposure"]["region"]}.')
                # ctrl.setAutoExposureRegion(*controls['exposure']['region'])
                LOGGER.error('exposure region not yet implemented')
            if controls['exposure'].get('compensation', None) is not None:
                exp_comp = clamp(controls['exposure']['compensation'], *LIMITS['exposure_compensation'])
                LOGGER.info(f'Setting exposure compensation to {exp_comp}.')
                ctrl.setAutoExposureCompensation(exp_comp)
            if controls['exposure'].get('manual', None) is not None:
                exp_time = clamp(controls['exposure']['manual'][0], *LIMITS['exposure'])
                exp_gain = clamp(controls['exposure']['manual'][1], *LIMITS['gain'])
                LOGGER.info(f'Setting exposure to {exp_time}, gain to {exp_gain}.')
                ctrl.setManualExposure(exp_time, exp_gain)

        if controls.get('ab_mode', None) is not None:
            if isinstance(controls["ab_mode"], str):
                controls["ab_mode"] = getattr(dai.CameraControl.AntiBandingMode, controls["ab_mode"])
            LOGGER.info(f'Setting anti-banding mode to {controls["ab_mode"]}.')
            ctrl.setAntiBandingMode(controls["ab_mode"])

        if controls.get('focus', None) is not None:
            if controls['focus'].get('range', None) is not None:
                LOGGER.info(f'Setting focus range to {controls["focus"]["range"]}.')
                ctrl.setAutoFocusLensRange(*controls['focus']['range'])
            if controls['focus'].get('region', None) is not None:
                # LOGGER.info(f'Setting focus region to {controls["focus"]["region"]}.')
                # ctrl.setAutoFocusRegion(*controls['focus']['region'])
                LOGGER.error('focus region not yet implemented')
            if controls['focus'].get('trigger', None) is not None:
                if controls['focus']['trigger']:
                    LOGGER.info('Auto-focus triggered.')
                    ctrl.setAutoFocusTrigger()
            if controls['focus'].get('mode', None) is not None:
                if isinstance(controls["focus"]["mode"], str):
                    controls["focus"]["mode"] = getattr(dai.CameraControl.AutoFocusMode, controls["focus"]["mode"])
                LOGGER.info(f'Setting focus mode to {controls["focus"]["mode"]}.')
                ctrl.setAutoFocusMode(controls["focus"]["mode"])
            if controls['focus'].get('manual', None) is not None:
                focus = clamp(controls['focus']['manual'], *LIMITS['focus'])
                LOGGER.info(f'Setting focus to {focus}.')
                ctrl.setManualFocus(focus)

        if controls.get('white-balance', None) is not None:
            if controls['white-balance'].get('mode', None) is not None:
                if isinstance(controls["focus"]["mode"], str):
                    controls["white-balance"]["mode"] = getattr(dai.CameraControl.AutoFocusMode,
                                                                controls["white-balance"]["mode"])
                LOGGER.info(f'Setting white balance mode to {controls["white-balance"]["mode"]}.')
                ctrl.setAutoWhiteBalanceMode(controls["white-balance"]["mode"])
            if controls['white-balance'].get('lock', None) is not None:
                LOGGER.info(f'Setting white balance lock to {controls["white-balance"]["lock"]}.')
                ctrl.setAutoWhiteBalanceLock(controls['white-balance']['lock'])
            if controls['white-balance'].get('manual', None) is not None:
                wb_temp = clamp(controls['white-balance']['manual'], *LIMITS['white_balance'])
                LOGGER.info(f'Setting white balance to {wb_temp}.')
                ctrl.setManualWhiteBalance(wb_temp)

        if controls.get('isp', None) is not None:
            if controls['isp'].get('brightness', None) is not None:
                brightness = clamp(controls['isp']['brightness'], *LIMITS['brightness'])
                LOGGER.info(f'Setting brightness to {brightness}.')
                ctrl.setBrightness(brightness)
            if controls['isp'].get('contrast', None) is not None:
                contrast = clamp(controls['isp']['contrast'], *LIMITS['contrast'])
                LOGGER.info(f'Setting contrast to {contrast}.')
                ctrl.setContrast(contrast)
            if controls['isp'].get('saturation', None) is not None:
                saturation = clamp(controls['isp']['saturation'], *LIMITS['saturation'])
                LOGGER.info(f'Setting saturation to {saturation}.')
                ctrl.setSaturation(saturation)
            if controls['isp'].get('sharpness', None) is not None:
                sharpness = clamp(controls['isp']['sharpness'], *LIMITS['sharpness'])
                LOGGER.info(f'Setting sharpness to {sharpness}.')
                ctrl.setSharpness(sharpness)
            if controls['isp'].get('luma_denoise', None) is not None:
                luma_denoise = clamp(controls['isp']['luma_denoise'], *LIMITS['luma_denoise'])
                LOGGER.info(f'Setting luma denoise to {luma_denoise}.')
                ctrl.setLumaDenoise(luma_denoise)
            if controls['isp'].get('chroma_denoise', None) is not None:
                chroma_denoise = clamp(controls['isp']['chroma_denoise'], *LIMITS['chroma_denoise'])
                LOGGER.info(f'Setting chroma denoise to {chroma_denoise}.')
                ctrl.setChromaDenoise(chroma_denoise)

        if controls.get('still', None) is not None:
            LOGGER.info('Capturing still photo.')
            ctrl.setCaptureStill(controls['still'])

        self.queue.send(ctrl)



from pathlib import Path
from string import Template
from typing import Union, Tuple

import depthai as dai

from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.components.camera_helper import get_resolution_size


def control_camera_with_nn(
        pipeline: dai.Pipeline,
        camera_control: dai.Node.Input,
        nn_output: dai.Node.Output,
        resize_mode: ResizeMode,
        resolution: Union[dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution],
        nn_size: Tuple[int, int],
        af: bool,
        ae: bool,
        debug: bool
):
    sensor_resolution = get_resolution_size(resolution)
    # width / height (old ar)
    sensor_ar = sensor_resolution[0] / sensor_resolution[1]
    # NN ar (new ar)
    nn_ar = nn_size[0] / nn_size[1]

    if resize_mode == ResizeMode.LETTERBOX:
        padding = (sensor_ar - nn_ar) / 2
        if padding > 0:
            init = f"xmin = 0; ymin = {-padding}; xmax = 1; ymax = {1 + padding}"
        else:
            init = f"xmin = {padding}; ymin = 0; xmax = {1 - padding}; ymax = 1"
    elif resize_mode in [ResizeMode.CROP, ResizeMode.FULL_CROP]:
        cropping = (1 - (nn_ar / sensor_ar)) / 2
        if cropping < 0:
            init = f"xmin = 0; ymin = {-cropping}; xmax = 1; ymax = {1 + cropping}"
        else:
            init = f"xmin = {cropping}; ymin = 0; xmax = {1 - cropping}; ymax = 1"
    else:  # Stretch
        init = f"xmin=0; ymin=0; xmax=1; ymax=1"

    resize_str = f"new_xmin=xmin+width*det.xmin; new_ymin=ymin+height*det.ymin; new_xmax=xmin+width*det.xmax; new_ymax=ymin+height*det.ymax;"
    denormalize = f"startx=int(new_xmin*{sensor_resolution[0]}); starty=int(new_ymin*{sensor_resolution[1]}); new_width=int((new_xmax-new_xmin)*{sensor_resolution[0]}); new_height=int((new_ymax-new_ymin)*{sensor_resolution[1]});"
    control_str = ''
    if ae:
        control_str += f"control.setAutoExposureRegion(startx, starty, new_width, new_height);"
    if af:
        control_str += f"control.setAutoFocusRegion(startx, starty, new_width, new_height);"

    script_node = pipeline.create(dai.node.Script)
    script_node.setProcessor(dai.ProcessorType.LEON_CSS)  # More stable

    with open(Path(__file__).parent / 'template_control_cam_with_nn.py', 'r') as file:
        code = Template(file.read()).substitute(
            DEBUG='' if debug else '#',
            INIT=init,
            RESIZE=resize_str,
            DENORMALIZE=denormalize,
            CONTROL=control_str
        )
        script_node.setScript(code)

        if debug:
            print(code)

    # Node linking:
    # NN output -> Script -> Camera input
    nn_output.link(script_node.inputs['detections'])
    script_node.outputs['control'].link(camera_control)



import json
from dataclasses import replace
from typing import List, Tuple, Optional, Union, Any, Dict

import depthai as dai
import numpy as np
from depthai import ImgDetection

from depthai_sdk.fps import FPSHandler
from depthai_sdk.visualize.bbox import BoundingBox
from depthai_sdk.visualize.configs import VisConfig, TextPosition, BboxStyle, StereoColor
from depthai_sdk.visualize.encoder import JSONEncoder
from depthai_sdk.visualize.objects import VisDetections, GenericObject, VisText, VisTrail, VisCircle, VisLine, VisMask, \
    VisBoundingBox


class VisualzierFps:
    def __init__(self):
        self.fps_list: Dict[str, FPSHandler] = {}

    def get_fps(self, name: str) -> float:
        if name not in self.fps_list:
            self.fps_list[name] = FPSHandler()

        self.fps_list[name].nextIter()
        return self.fps_list[name].fps()


class Visualizer:
    # Constants
    def __init__(self, scale: float = None, fps: bool = False):
        self.objects: List[GenericObject] = []
        self._frame_shape: Optional[Tuple[int, ...]] = None

        self.config = VisConfig()
        self.fps = VisualzierFps()

        if fps:
            self.output(show_fps=fps)
        if scale:
            self.output(img_scale=float(scale))

    def add_object(self, obj: GenericObject) -> 'Visualizer':
        """
        Call `set_config`, `set_frame_shape` and `prepare` for the object and add it to the list of objects.
        Args:
            obj: The object to add.

        Returns:
            self
        """
        obj = obj.set_config(self.config).set_frame_shape(self.frame_shape).prepare()
        self.objects.append(obj)
        return self

    def add_bbox(self,
                 bbox: BoundingBox,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None,
                 bbox_style: BboxStyle = None,
                 label: str = None,
                 ) -> 'Visualizer':
        """
        Add a bounding box to the visualizer.

        Args:
            bbox: Bounding box.
            label: Label for the detection.
            thickness: Bounding box thickness.
            color: Bounding box color (RGB).
            bbox_style: Bounding box style (one of depthai_sdk.visualize.configs.BboxStyle).

        Returns:
            self
        """
        bbox_overlay = VisBoundingBox(bbox=bbox,
                                      color=color,
                                      thickness=thickness,
                                      bbox_style=bbox_style,
                                      label=label)
        self.add_object(bbox_overlay)
        return self

    def add_detections(self,
                       detections: List[Union[ImgDetection, dai.Tracklet]],
                       normalizer: BoundingBox = None,
                       label_map: List[Tuple[str, Tuple]] = None,
                       spatial_points: List[dai.Point3f] = None,
                       label_color: Tuple[int, int, int] = None,
                       label_background_color: Tuple[int, int, int] = None,
                       label_background_transparency: float = None,
                       is_spatial=False,
                       bbox: Union[np.ndarray, Tuple[int, int, int, int]] = None,
                       ) -> 'Visualizer':
        """
        Add detections to the visualizer.

        Args:
            detections: List of detections.
            normalizer: Normalizer object.
            label_map: List of tuples (label, color).
            spatial_points: List of spatial points. None if not spatial.
            label_color: Color for the label.
            label_background_color: Color for the label background.
            label_background_transparency: Transparency for the label background.
            is_spatial: Flag that indicates if the detections are spatial.
            bbox: Bounding box, if there's a detection inside a bounding box.

        Returns:
            self
        """
        detection_overlay = VisDetections(detections=detections,
                                          normalizer=normalizer,
                                          label_map=label_map,
                                          spatial_points=spatial_points,
                                          label_color=label_color,
                                          label_background_color=label_background_color,
                                          label_background_transparency=label_background_transparency,
                                          is_spatial=is_spatial,
                                          parent_bbox=bbox)
        self.add_object(detection_overlay)
        return self

    def add_text(self,
                 text: str,
                 coords: Tuple[int, int] = None,
                 size: int = None,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None,
                 outline: bool = True,
                 background_color: Tuple[int, int, int] = None,
                 background_transparency: float = 0.5,
                 bbox: Union[np.ndarray, Tuple, BoundingBox] = None,
                 position: TextPosition = TextPosition.TOP_LEFT,
                 padding: int = 10) -> 'Visualizer':
        """
        Add text to the visualizer.

        Args:
            text: Text.
            coords: Coordinates.
            size: Size of the text.
            color: Color of the text.
            thickness: Thickness of the text.
            outline: Flag that indicates if the text should be outlined.
            background_color: Background color.
            background_transparency: Background transparency.
            bbox: Bounding box.
            position: Position.
            padding: Padding.

        Returns:
            self
        """
        if isinstance(bbox, Tuple) and type(bbox[0]) == float:
            bbox = BoundingBox(bbox)

        text_overlay = VisText(text=text,
                               coords=coords,
                               size=size,
                               color=color,
                               thickness=thickness,
                               outline=outline,
                               background_color=background_color,
                               background_transparency=background_transparency,
                               bbox=bbox,
                               position=position,
                               padding=padding)
        self.add_object(text_overlay)
        return self

    def add_trail(self,
                  tracklets: List[dai.Tracklet],
                  label_map: List[Tuple[str, Tuple]],
                  bbox: BoundingBox = None) -> 'Visualizer':
        """
        Add a trail to the visualizer.

        Args:
            tracklets: List of tracklets.
            label_map: List of tuples (label, color).
            bbox: Bounding box.

        Returns:
            self
        """
        if bbox is None:
            bbox = BoundingBox()

        trail = VisTrail(tracklets=tracklets,
                         label_map=label_map,
                         bbox=bbox)
        self.add_object(trail)
        return self

    def add_circle(self,
                   coords: Tuple[int, int],
                   radius: int,
                   color: Tuple[int, int, int] = None,
                   thickness: int = None) -> 'Visualizer':
        """
        Add a circle to the visualizer.

        Args:
            coords: Center of the circle.
            radius: Radius of the circle.
            color: Color of the circle.
            thickness: Thickness of the circle.

        Returns:
            self
        """
        circle = VisCircle(coords=coords,
                           radius=radius,
                           color=color,
                           thickness=thickness)
        self.objects.append(circle)
        return self

    def add_line(self,
                 pt1: Tuple[int, int],
                 pt2: Tuple[int, int],
                 color: Tuple[int, int, int] = None,
                 thickness: int = None) -> 'Visualizer':
        """
        Add a line to the visualizer.

        Args:
            pt1: Start coordinates.
            pt2: End coordinates.
            color: Color of the line.
            thickness: Thickness of the line.

        Returns:
            self
        """
        line = VisLine(pt1=pt1,
                       pt2=pt2,
                       color=color,
                       thickness=thickness)
        self.objects.append(line)
        return self

    def add_mask(self, mask: np.ndarray, alpha: float):
        """
        Add a mask to the visualizer.

        Args:
            mask: Mask represented as uint8 numpy array.
            alpha: Transparency of the mask.

        Returns:
            self
        """
        mask_overlay = VisMask(mask=mask, alpha=alpha)
        self.add_object(mask_overlay)
        return self

    def drawn(self, frame: np.ndarray) -> Optional[np.ndarray]:
        """
        Draw all objects on the frame if the platform is PC. Otherwise, serialize the objects
        and communicate with the RobotHub application.

        Args:
            frame: The frame to draw on.

        Returns:
            np.ndarray if the platform is PC, None otherwise.
        """
        raise NotImplementedError('Visualizers that inherit from Visualizer must implement draw() method!')

    def show(self, packet):
        """
        Show the packet on the screen.
        """
        pass

    def serialize(self, force_reset: bool = True) -> str:
        """
        Serialize all contained objects to JSON.

        Args:
            force_reset: Flag that indicates if the objects should be cleared after serialization.

        Returns:
            Stringified JSON.
        """
        parent = {
            'frame_shape': self.frame_shape,
            'config': self.config,
            'objects': [obj.serialize() for obj in self.objects]
        }

        if force_reset:
            self.reset()

        return json.dumps(parent, cls=JSONEncoder)

    def reset(self):
        """Clear all objects."""
        self.objects.clear()

    def output(self,
               img_scale: float = None,
               show_fps: bool = None) -> 'Visualizer':
        """
        Configure the output of the visualizer.

        Args:
            img_scale: Scale of the output image.
            show_fps: Flag that indicates if the FPS should be shown.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.output = replace(self.config.output, **kwargs)

        return self

    def stereo(self,
               colorize: StereoColor = None,
               colormap: int = None,
               wls_filter: bool = None,
               wls_lambda: float = None,
               wls_sigma: float = None,
               depth_score: bool = None):
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.stereo = replace(self.config.stereo, **kwargs)

        return self

    def detections(self,
                   thickness: int = None,
                   fill_transparency: float = None,
                   bbox_roundness: float = None,
                   color: Tuple[int, int, int] = None,
                   bbox_style: BboxStyle = None,
                   line_width: float = None,
                   line_height: float = None,
                   hide_label: bool = None,
                   label_position: TextPosition = None,
                   label_padding: int = None) -> 'Visualizer':
        """
        Configure how bounding boxes will look like.
        Args:
            thickness: Thickness of the bounding box.
            fill_transparency: Transparency of the bounding box.
            bbox_roundness: Roundness of the bounding box.
            color: Color of the bounding box.
            bbox_style: Style of the bounding box.
            line_width: Width of the bbox horizontal lines CORNERS or ROUNDED_CORNERS style is used.
            line_height: Height of the bbox vertical lines when CORNERS or ROUNDED_CORNERS style is used.
            hide_label: Flag that indicates if the label should be hidden.
            label_position: Position of the label.
            label_padding: Padding of the label.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.detection = replace(self.config.detection, **kwargs)

        return self

    def text(self,
             font_face: int = None,
             font_color: Tuple[int, int, int] = None,
             font_transparency: float = None,
             font_scale: float = None,
             font_thickness: int = None,
             font_position: TextPosition = None,
             background_transparency: float = None,
             background_color: Tuple[int, int, int] = None,
             outline_color: Tuple[int, int, int] = None,
             line_type: int = None,
             auto_scale: bool = None) -> 'Visualizer':
        """
        Configure how text will look like.

        Args:
            font_face: Font face (from cv2).
            font_color: Font color.
            font_transparency: Font transparency.
            font_scale: Font scale.
            font_thickness: Font thickness.
            font_position: Font position.
            background_transparency: Text background transparency.
            background_color: Text background color.
            outline_color: Outline color.
            line_type: Line type (from cv2).
            auto_scale: Flag that indicates if the font scale should be automatically adjusted.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.text = replace(self.config.text, **kwargs)

        return self

    def tracking(self,
                 max_length: int = None,
                 deletion_lost_threshold: int = None,
                 line_thickness: int = None,
                 fading_tails: bool = None,
                 show_speed: bool = None,
                 line_color: Tuple[int, int, int] = None,
                 line_type: int = None,
                 bg_color: Tuple[int, int, int] = None) -> 'Visualizer':
        """
        Configure how tracking trails will look like.

        Args:
            max_length: Maximum length of the trail (in pixels).
            deletion_lost_threshold: Number of consequent LOST statuses after which the trail is deleted.
            line_thickness: Thickness of the line.
            fading_tails: Flag that indicates if the tails should fade.
            show_speed: Flag that indicates if the speed should be shown.
            line_color: Color of the line.
            line_type: Type of the line (from cv2).
            bg_color: Text background color.

        Returns:
            self
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.tracking = replace(self.config.tracking, **kwargs)

        return self

    def segmentation(self,
                     mask_alpha: float = None,
                     ) -> 'Visualizer':
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.config.segmentation = replace(self.config.segmentation, **kwargs)

        return self

    @property
    def frame_shape(self) -> Tuple[int, ...]:
        return self._frame_shape

    @frame_shape.setter
    def frame_shape(self, shape: Tuple[int, ...]) -> None:
        self._frame_shape = shape

    @staticmethod
    def _process_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Process the kwargs and remove all None values."""
        kwargs.pop('self')
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return kwargs

    def close(self):
        pass



import subprocess
import sys

import depthai_viewer as viewer
import numpy as np
from depthai_viewer.components.rect2d import RectFormat

from depthai_sdk.classes.packets import FramePacket, IMUPacket, PointcloudPacket
from depthai_sdk.logger import LOGGER
from depthai_sdk.visualize.objects import (
    VisBoundingBox,
    VisCircle,
    VisDetections,
    VisLine,
    VisMask,
    VisText,
    VisTrail,
)
from depthai_sdk.visualize.visualizer import Visualizer


class DepthaiViewerVisualizer(Visualizer):
    """
    Visualizer for Depthai Viewer (https://github.com/luxonis/depthai-viewer)
    """

    def __init__(self, scale, fps):
        super().__init__(scale, fps)

        try:
            # timeout is optional, but it might be good to prevent the script from hanging if the module is large.
            process = subprocess.Popen([sys.executable, "-m", "depthai_viewer", "--viewer-mode"], stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            stdout, stderr = process.communicate(timeout=3)

            if process.returncode != 0:
                err_msg = stderr.decode("utf-8")
                if 'Failed to bind TCP address' in err_msg:
                    # Already running
                    pass
                elif 'No module named depthai_viewer' in err_msg:
                    raise Exception(f"DepthAI Viewer is not installed. "
                                    f"Please run '{sys.executable} -m pip install depthai_viewer' to install it.")
                else:
                    LOGGER.exception(f"Error occurred while trying to run depthai_viewer: {err_msg}")
            else:
                print("depthai_viewer ran successfully.")
        except subprocess.TimeoutExpired:
            # Installed and running depthai_viewer successfully
            pass
        except subprocess.CalledProcessError as e:
            print(f"An error occurred while trying to run 'depthai_viewer': {str(e)}")

        viewer.init("Depthai Viewer")
        viewer.connect()

    def show(self, packet) -> None:
        if isinstance(packet, FramePacket):
            bgr_frame = packet.decode()
            rgb_frame = bgr_frame[..., ::-1]
            frame = np.dstack((rgb_frame, np.full(bgr_frame.shape[:2], 255, dtype=np.uint8)))
            viewer.log_image(packet.name, frame)

        if type(packet) == IMUPacket:
            viewer.log_imu(*packet.get_imu_vals())
        elif type(packet) == PointcloudPacket:
            if packet.colorize_frame is not None:
                rgb_frame = packet.colorize_frame[..., ::-1]
                viewer.log_image(f'color', rgb_frame)
                viewer.log_points(packet.name, packet.points.reshape(-1, 3) / 1000, colors=rgb_frame.reshape(-1, 3))
            else:
                viewer.log_points(packet.name, packet.points.reshape(-1, 3) / 1000)

        vis_bbs = []
        for i, obj in enumerate(self.objects):
            if type(obj) == VisBoundingBox:
                vis_bbs.append(obj)
            elif type(obj) == VisDetections:
                pass
            elif type(obj) == VisText:
                pass
            elif type(obj) == VisTrail:
                pass
            elif type(obj) == VisLine:
                pass
            elif type(obj) == VisCircle:
                pass
            elif type(obj) == VisMask:
                pass

        if 0 < len(vis_bbs):
            rects = [vis_bb.bbox.clip().denormalize(frame.shape) for vis_bb in vis_bbs]
            # Convert from (pt1,pt2) to [x1,y1,x2,y2]
            rects = [np.array([*rect[0], *rect[1]]) for rect in rects]
            # BGR to RGB
            colors = [np.array(vis_bb.color)[..., ::-1] for vis_bb in vis_bbs]
            labels = [vis_bb.label for vis_bb in vis_bbs]
            print(rects)
            viewer.log_rects(
                f"{packet.name}/Detections",
                rects=rects,
                rect_format=RectFormat.XYXY,
                colors=colors,
                labels=labels
            )
        self.reset()

    def close(self):
        pass



from abc import ABC, abstractmethod
from typing import List, Optional, Callable

import depthai as dai

from depthai_sdk.classes.packets import FramePacket
from depthai_sdk.components.component import ComponentOutput


class StreamXout:
    def __init__(self, out: dai.Node.Output, name: Optional[str] = None):
        self.stream = out
        if name is not None:
            self.name = name
        else:
            node = out.getParent()
            self.name = f"{str(node.id)}_{out.name}"


class ReplayStream(StreamXout):
    def __init__(self, name: str):
        self.name = name


class XoutBase(ABC):
    def __init__(self) -> None:
        self._streams = [xout.name for xout in self.xstreams()]
        self._packet_name = None
        self._packet_name_postfix = None

        # It will get assigned later inside the BasePacketHandler class
        self.new_packet_callback: Callable = lambda x: None

    def get_packet_name(self) -> str:
        if self._packet_name is None:
            self._packet_name = ";".join([xout.name for xout in self.xstreams()])
        return self._packet_name + (f'_{self._packet_name_postfix}' if self._packet_name_postfix else '')

    def set_packet_name_postfix(self, postfix: str) -> None:
        """
        Set postfix to packet name.
        """
        self._packet_name_postfix = postfix

    def set_comp_out(self, comp_out: ComponentOutput) -> 'XoutBase':
        """
        Set ComponentOutput to Xout.
        """
        if comp_out.name is None:
            # If user hasn't specified component's output name, generate one
            comp_out.name = self.get_packet_name()
        else:
            # Otherwise, set packet name to user-specified one
            self._packet_name = comp_out.name
        return self

    @abstractmethod
    def xstreams(self) -> List[StreamXout]:
        raise NotImplementedError()

    def device_msg_callback(self, name, dai_message) -> None:
        """
        This is the (first) callback that gets called on a device message. Don't override it.
        It will call `new_msg` and `on_callback` methods. If `new_msg` returns a packet, it will call
        `new_packet` method.
        """
        # self._fps_counter[name].next_iter()
        packet = self.new_msg(name, dai_message)
        if packet is not None:
            # If not list, convert to list.
            # Some Xouts create multiple packets from a single message (example: IMU)
            if not isinstance(packet, list):
                packet = [packet]

            for p in packet:
                # In case we have encoded frames, we need to set the codec
                if isinstance(p, FramePacket) and \
                        hasattr(self, 'get_codec') and \
                        self._fourcc is not None:
                    p.set_decode_codec(self.get_codec)

                self.on_callback(p)
                self.new_packet_callback(p)

    @abstractmethod
    def new_msg(self, name: str, msg) -> None:
        raise NotImplementedError()

    def on_callback(self, packet) -> None:
        """
        Hook called when `callback` or `self.visualize` are used.
        """
        pass

    def fourcc(self) -> str:
        if self.is_mjpeg():
            return 'mjpeg'
        elif self.is_h264():
            return 'h264'
        elif self.is_h265():
            return 'hevc'
        elif self.is_depth():
            return 'y16'
        # TODO: add for mono, rgb, nv12, yuv...
        else:
            return None

    def is_h265(self) -> bool:
        fourcc = getattr(self, '_fourcc', None)
        return fourcc is not None and fourcc.lower() == 'hevc'

    def is_h264(self) -> bool:
        fourcc = getattr(self, '_fourcc', None)
        return fourcc is not None and fourcc.lower() == 'h264'

    def is_mjpeg(self) -> bool:
        fourcc = getattr(self, '_fourcc', None)
        return fourcc is not None and fourcc.lower() == 'mjpeg'

    def is_h26x(self) -> bool:
        return self.is_h264() or self.is_h265()

    def is_raw(self) -> bool:
        return type(self).__name__ == 'XoutFrames' and self._fourcc is None

    def is_depth(self) -> bool:
        return type(self).__name__ == 'XoutDepth'

    def is_disparity(self) -> bool:
        return type(self).__name__ == 'XoutDisparity'

    def is_imu(self):
        return type(self).__name__ == 'XoutIMU'



import math
from datetime import timedelta
from typing import Dict, Optional, List, Union, Tuple

import depthai as dai
import numpy as np

from depthai_sdk.classes import TrackerPacket
from depthai_sdk.classes.packets import TrackingDetection
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_nn import XoutNnResults
from depthai_sdk.tracking import KalmanFilter
from depthai_sdk.visualize.bbox import BoundingBox


class TrackedObject:
    def __init__(self, baseline: float, focal: float, apply_kalman: bool, calculate_speed: bool):
        # Point
        self.kalman_3d: Optional[KalmanFilter] = None
        # BBox
        self.kalman_2d: Optional[KalmanFilter] = None

        self.previous_detections: List[TrackingDetection] = []
        self.blacklist = False
        self.lost_counter = 0

        self.baseline = baseline
        self.focal = focal
        self.apply_kalman = apply_kalman
        self.calculate_speed = calculate_speed

    def new_tracklet(self, tracklet: dai.Tracklet, ts: timedelta, color: Tuple, label: str):
        is_3d = self._is_3d(tracklet)
        tracking_det = TrackingDetection(
            img_detection=tracklet.srcImgDetection,
            label_str=label,
            confidence=tracklet.srcImgDetection.confidence,
            color=color,
            bbox=BoundingBox(tracklet.srcImgDetection),
            angle=None,
            tracklet=tracklet,
            ts=ts,
            filtered_2d=self._calc_kalman_2d(tracklet, ts) if self.apply_kalman else None,
            filtered_3d=self._calc_kalman_3d(tracklet, ts) if self.apply_kalman and is_3d else None,
            speed=None,
        )
        self.previous_detections.append(tracking_det)
        # Calc speed should be called after adding new TrackingDetection to self.previous_detections
        tracking_det.speed = self.calc_speed(ts) if (self.calculate_speed and is_3d) else None

    def calc_speed(self, ts: timedelta) -> Union[float, np.ndarray]:
        """
        Should be called after adding new TrackingDetection to self.previous_detections
        """

        def get_coords(det) -> dai.Point3f:
            return det.filtered_3d or det.tracklet.spatialCoordinates

        def get_dist(p1: dai.Point3f, p2: dai.Point3f) -> float:
            return np.sqrt((p2.x - p1.x) ** 2 + (p2.y - p1.y) ** 2 + (p2.z - p1.z) ** 2) / 1000

        speeds = []
        for i in range(len(self.previous_detections) - 1):
            d1 = self.previous_detections[i]
            # if d1 timestamp is older than 1 second, skip
            if (ts - d1.ts).total_seconds() > 1:
                continue
            d2 = self.previous_detections[i + 1]
            distance = get_dist(get_coords(d1), get_coords(d2))
            time = (d2.ts - d1.ts).total_seconds()
            speeds.append(distance / time)

        if len(speeds) == 0:
            return 0.0

        window_size = 3
        window = np.hanning(window_size)
        window /= window.sum()

        smoothed = np.convolve(speeds, window, mode='same')
        return np.mean(smoothed)

    def _is_3d(self, tracklet: dai.Tracklet) -> bool:
        return (tracklet.spatialCoordinates.x != 0.0 or
                tracklet.spatialCoordinates.y != 0.0 or
                tracklet.spatialCoordinates.z != 0.0)

    def _calc_kalman_3d(self, tracklet: dai.Tracklet, ts: timedelta) -> Union[None, dai.Point3f]:
        x_space = tracklet.spatialCoordinates.x
        y_space = tracklet.spatialCoordinates.y
        z_space = tracklet.spatialCoordinates.z
        meas_vec_space = np.array([[x_space], [y_space], [z_space]])
        meas_std_space = z_space ** 2 / (self.baseline * self.focal)

        if self.kalman_3d is None:
            self.kalman_3d = KalmanFilter(10, 0.1, meas_vec_space, ts)
            return None

        dt = (ts - self.kalman_3d.time).total_seconds()
        self.kalman_3d.predict(dt)
        self.kalman_3d.update(meas_vec_space)
        self.kalman_3d.time = ts
        self.kalman_3d.meas_std = meas_std_space
        vec_space = self.kalman_3d.x
        return dai.Point3f(vec_space[0], vec_space[1], vec_space[2])

    def _calc_kalman_2d(self, tracklet: dai.Tracklet, ts: timedelta) -> Union[None, BoundingBox]:
        bb = BoundingBox(tracklet.srcImgDetection)
        x_mid, y_mid = bb.get_centroid().to_tuple()

        meas_vec_bbox = np.array([[x_mid], [y_mid], [bb.width], [bb.height]])

        if self.kalman_2d is None:
            self.kalman_2d = KalmanFilter(10, 0.1, meas_vec_bbox, ts)
            return None

        dt = (ts - self.kalman_2d.time).total_seconds()

        self.kalman_2d.predict(dt)
        self.kalman_2d.update(meas_vec_bbox)
        self.kalman_2d.time = ts
        vec_bbox = self.kalman_2d.x

        return BoundingBox([
            vec_bbox[0][0] - vec_bbox[2][0] / 2,
            vec_bbox[1][0] - vec_bbox[3][0] / 2,
            vec_bbox[0][0] + vec_bbox[2][0] / 2,
            vec_bbox[1][0] + vec_bbox[3][0] / 2,
        ])


class XoutTracker(XoutNnResults):
    def __init__(self,
                 det_nn: 'NNComponent',
                 frames: StreamXout,
                 device: dai.Device,
                 tracklets: StreamXout,
                 bbox: BoundingBox,
                 apply_kalman: bool = False,
                 forget_after_n_frames: Optional[int] = None,
                 calculate_speed: bool = False,
                 ):
        """
        apply_kalman: Whether to apply kalman filter to tracklets
        forget_after_n_frames: If tracklet is lost for n frames, remove it from tracked_objects

        """
        super().__init__(det_nn, frames, tracklets, bbox)
        self.name = 'Object Tracker'
        self.__read_device_calibration(device)

        self.tracked_objects: Dict[int, TrackedObject] = {}

        self.apply_kalman = apply_kalman
        self.forget_after_n_frames = forget_after_n_frames
        self.calculate_speed = calculate_speed

    def package(self, msgs: Dict) -> TrackerPacket:
        tracklets: dai.Tracklets = msgs[self.nn_results.name]

        for tracklet in tracklets.tracklets:
            # If there is no id in self.tracked_objects, create new TrackedObject. This could happen if
            # TrackingStatus.NEW, or we removed it (too many lost frames)
            if tracklet.id not in self.tracked_objects:
                self.tracked_objects[tracklet.id] = TrackedObject(self.baseline, self.focal, self.apply_kalman,
                                                                  self.calculate_speed)

            if tracklet.status == dai.Tracklet.TrackingStatus.NEW:
                pass
            elif tracklet.status == dai.Tracklet.TrackingStatus.TRACKED:
                self.tracked_objects[tracklet.id].lost_counter = 0
            elif tracklet.status == dai.Tracklet.TrackingStatus.LOST:
                self.tracked_objects[tracklet.id].lost_counter += 1

            img_d = tracklet.srcImgDetection
            # When adding new tracklet, TrackletObject class will also perform filtering
            # and speed estimation
            self.tracked_objects[tracklet.id] \
                .new_tracklet(tracklet,
                              tracklets.getTimestamp(),
                              self.labels[img_d.label][1] if self.labels else (255, 255, 255),
                              self.labels[img_d.label][0] if self.labels else str(img_d.label)
                              )
            if tracklet.status == dai.Tracklet.TrackingStatus.REMOVED or \
                    (self.forget_after_n_frames is not None and \
                     self.forget_after_n_frames <= self.tracked_objects[tracklet.id].lost_counter):
                # Remove TrackedObject
                self.tracked_objects.pop(tracklet.id)

        packet = TrackerPacket(
            self.get_packet_name(),
            msgs[self.frames.name],
            tracklets,
            bbox=self.bbox,
        )

        for obj_id, tracked_obj in self.tracked_objects.items():
            if obj_id not in packet.tracklets:
                packet.tracklets[obj_id] = []
            for tracking_det in tracked_obj.previous_detections:
                packet.tracklets[obj_id].append(tracking_det)

        return packet

    def __read_device_calibration(self, device: dai.Device):
        calib = device.readCalibration()
        eeprom = calib.getEepromData()
        left_cam = calib.getStereoLeftCameraId()
        if left_cam != dai.CameraBoardSocket.AUTO and left_cam in eeprom.cameraData.keys():
            cam_info = eeprom.cameraData[left_cam]
            self.baseline = abs(cam_info.extrinsics.specTranslation.x * 10)  # cm -> mm
            fov = calib.getFov(calib.getStereoLeftCameraId())
            self.focal = (cam_info.width / 2) / (2. * math.tan(math.radians(fov / 2)))
        else:
            LOGGER.warning("Calibration data missing, using OAK-D defaults")
            self.baseline = 75
            self.focal = 440



from abc import ABC
from typing import Callable, Union, List, Dict, Optional

from depthai_sdk.classes import FramePacket
from depthai_sdk.components import Component

__all__ = ['Action']


class Action(ABC):
    """
    Base action represents a single action that can be activated by a trigger.
    """

    def __init__(self,
                 inputs: Optional[Union[Component, Callable, List[Union[Component, Callable]]]] = None,
                 action: Optional[Callable] = None):
        """
        Args:
            inputs: Single or multiple input components or outputs of components that will be used as input for the action.
            action: Action that will be executed when the trigger is activated. Should be a callable object.
        """
        if inputs:
            if not isinstance(inputs, list):
                inputs = [inputs]

            for i in range(len(inputs)):
                if isinstance(inputs[i], Component):
                    inputs[i] = inputs[i].out.main

        self.inputs = inputs
        self.stream_names: List[str] = []
        self.action = action

    def activate(self) -> None:
        """
        Method that gets called when the action is activated by a trigger.
        """
        if self.action:
            self.action()

    def on_new_packets(self, packets: Dict[str, FramePacket]) -> None:
        """
        Callback method that gets called when all packets are synced.

        Args:
            packets: Dictionary of packets received from the input streams.
        """
        pass



import depthai as dai
from itertools import cycle
from depthai_sdk.components.parser import parse_median_filter
from depthai_sdk.logger import LOGGER

LIMITS = {
    'confidence_threshold': (0, 255),
    'bilateral_sigma': (0, 255),
    'range': (0, 65535),
    'lrc_threshold': (0, 10),
    'dot_projector': (0, 1200),
    'illumination_led': (0, 1500),
}


def clamp(value, min_value, max_value):
    return max(min(value, max_value), min_value)


class StereoControl:
    def __init__(self, device: dai.Device):
        self.queue = None

        self._cycle_median_filter = cycle([item for name, item in vars(dai.StereoDepthConfig.MedianFilter).items() if name[0].isupper()])
        self.device = device

        self._current_vals = {
            'conf_threshold': 240,
            'dot_projector': 800,
            'illumination_led': 200,
        }

        ctrl = dai.StereoDepthConfig()
        self.raw_cfg = ctrl.get()

    def set_input_queue(self, queue: dai.DataInputQueue):
        self.queue = queue

    def switch_median_filter(self):
        """
        Switch between auto white balance modes.
        """
        mode = next(self._cycle_median_filter)
        LOGGER.info(f'Switching median filter to {mode}')
        self.send_controls({'postprocessing': {'median': mode}})

    def confidence_threshold_up(self, step=10):
        """
        Increase confidence threshold by step
        """
        if LIMITS['confidence_threshold'][1] < self._current_vals['conf_threshold'] + step:
            LOGGER.error(f'Confidence threshold cannot be greater than {LIMITS["confidence_threshold"][1]}')
            return
        self._current_vals['conf_threshold'] += step
        self.send_controls({'cost_matching': {'confidence_threshold': self._current_vals['conf_threshold']}})

    def confidence_threshold_down(self, step=10):
        """
        Decrease confidence threshold by step
        """
        if LIMITS['confidence_threshold'][0] > self._current_vals['conf_threshold'] - step:
            LOGGER.error(f'Confidence threshold cannot be less than {LIMITS["confidence_threshold"][0]}')
            return
        self._current_vals['conf_threshold'] -= step
        self.send_controls({'cost_matching': {'confidence_threshold': self._current_vals['conf_threshold']}})

    def dot_projector_up(self, step=50):
        """
        Increase dot projector power by step
        """
        if LIMITS['dot_projector'][1] < self._current_vals['dot_projector'] + step:
            LOGGER.error(f'Dot projector power cannot be greater than {LIMITS["dot_projector"][1]}')
            return
        self._current_vals['dot_projector'] += step
        self.device.setIrFloodLightBrightness(self._current_vals['dot_projector'])

    def dot_projector_down(self, step=50):
        """
        Decrease dot projector power by step
        """
        if LIMITS['dot_projector'][0] > self._current_vals['dot_projector'] - step:
            LOGGER.error(f'Dot projector power cannot be less than {LIMITS["dot_projector"][0]}')
            return
        self._current_vals['dot_projector'] -= step
        self.device.setIrFloodLightBrightness(self._current_vals['dot_projector'])

    def illumination_led_up(self, step=50):
        """
        Increase illumination led power by step
        """
        if LIMITS['illumination_led'][1] < self._current_vals['illumination_led'] + step:
            LOGGER.error(f'Illumination led power cannot be greater than {LIMITS["illumination_led"][1]}')
            return
        self._current_vals['illumination_led'] += step
        self.device.setIrLaserDotProjectorBrightness(self._current_vals['illumination_led'])

    def illumination_led_down(self, step=50):
        """
        Decrease illumination led power by step
        """
        if LIMITS['illumination_led'][0] > self._current_vals['illumination_led'] - step:
            LOGGER.error(f'Illumination led power cannot be less than {LIMITS["illumination_led"][0]}')
            return
        self._current_vals['illumination_led'] -= step
        self.device.setIrLaserDotProjectorBrightness(self._current_vals['illumination_led'])

    def send_controls(self, controls: dict):
        """
        Send controls to the StereoDepth node. Dict structure and available options:

        ctrl = {
            'algorithm_control':{
                'align': 'RECTIFIED_RIGHT', # | 'RECTIFIED_LEFT' | 'CENTER'
                'unit': 'METER', # | 'CENTIMETER' | 'MILLIMETER' | 'INCH' | 'FOOT' | 'CUSTOM'
                'unit_multiplier': 1000, # Only if 'unit' is 'CUSTOM'
                'lr_check': True, # Enable left-right check
                'extended': True, # Enable extended disparity
                'subpixel': True, # Enable subpixel disparity
                'lr_check_threshold': 10, # Left-right check threshold
                'subpixel_bits': 3, # 3 | 4 | 5
                'disparity_shift': 0, # Disparity shift
                'invalidate_edge_pixels': 0 # Number of pixels to invalidate at the edge of the image
            },
            'postprocessing': {
                'median': 5, # 0 | 3 | 5 | 7
                'bilateral_sigma': 0, # Sigma value for bilateral filter
                'spatial': {
                    'enable': True, # Enable spatial denoise
                    'hole_filling': 2, # Hole filling radius
                    'alpha': 0.5, # Alpha factor in an exponential moving average
                    'delta': 0, # Step-size boundary
                    'iterations': 1, # Number of iterations
                },
                'temporal': {
                    'enable': False,  # Enable or disable temporal denoise
                    'persistency_mode': 3,  # Persistency mode (use corresponding integer value for the enum member)
                    'alpha': 0.4,  # Alpha factor in an exponential moving average
                    'delta': 0,  # Step-size boundary
                },
                'threshold': {
                    'min_range': 0,  # Minimum range in depth units
                    'max_range': 65535,  # Maximum range in depth units
                },
                'brightness': {
                    'min': 0,  # Minimum pixel brightness
                    'max': 256,  # Maximum pixel brightness
                },
                'speckle': {
                    'enable': False,  # Enable or disable the speckle filter
                    'range': 50,  # Speckle search range
                },
                'decimation': {
                    'factor': 1,  # Decimation factor (1, 2, 3, or 4)
                    'mode': 0,  # Decimation algorithm type (use corresponding integer value for the enum member)
                }
            },
            'census_transform': {
                'kernel_size': 'AUTO',  # | 'KERNEL_5x5' | 'KERNEL_7x7' | 'KERNEL_7x9'
                'kernel_mask': 0,  # Census transform mask
                'enable_mean_mode': True,  # Enable mean mode
                'threshold': 0,  # Census transform comparison threshold value
            },
            'cost_matching': {
                'disparity_width': 'DISPARITY_64',  # or 'DISPARITY_96'
                'enable_companding': False,  # Enable disparity companding using sparse matching
                'confidence_threshold': 245,  # Confidence threshold for accepted disparities
                'linear_equation_parameters': {
                    'alpha': 0,
                    'beta': 2,
                    'threshold': 127,
                },
            },
            'cost_aggregation': {
                'division_factor': 1,  # Division factor for cost calculation linear equation parameters
                'horizontal_penalty_cost_p1': 250,  # Horizontal P1 penalty cost parameter
                'horizontal_penalty_cost_p2': 500,  # Horizontal P2 penalty cost parameter
                'vertical_penalty_cost_p1': 250,  # Vertical P1 penalty cost parameter
                'vertical_penalty_cost_p2': 500,  # Vertical P2 penalty cost parameter
            },
            'reset': False # Reset all controls to default
        }
        """
        if self.queue is None:
            LOGGER.error('Cannot send controls when replaying.')
            return

        LOGGER.info(f'Sending controls to StereoDepth node: {controls}')

        ctrl = dai.StereoDepthConfig()

        if controls.get('reset', None) and controls['reset']:
            LOGGER.info('Resetting camera controls.')
            self.raw_cfg = ctrl.get()
            ctrl.set(self.raw_cfg)
            self.queue.send(ctrl)
            return

        if controls.get('algorithm_control', None) is not None:
            if controls['algorithm_control'].get('align', None) is not None:
                if isinstance(controls["algorithm_control"]["align"], str):
                    controls["algorithm_control"]["align"] = getattr(dai.StereoDepthConfig.AlgorithmControl.DepthAlign, controls["algorithm_control"]["align"])
                self.raw_cfg.algorithmControl.depthAlign = controls["algorithm_control"]["align"]
            if controls['algorithm_control'].get('unit', None) is not None:
                if isinstance(controls["algorithm_control"]["unit"], str):
                    controls["algorithm_control"]["unit"] = getattr(dai.StereoDepthConfig.AlgorithmControl.DepthUnit, controls["algorithm_control"]["unit"])
                self.raw_cfg.algorithmControl.depthUnit = controls["algorithm_control"]["unit"]
            if controls['algorithm_control'].get('unit_multiplier', None) is not None:
                self.raw_cfg.algorithmControl.customDepthUnitMultiplier = controls["algorithm_control"]["unit_multiplier"]
            if controls['algorithm_control'].get('lr_check', None) is not None:
                self.raw_cfg.algorithmControl.enableLeftRightCheck = controls["algorithm_control"]["lr_check"]
            if controls['algorithm_control'].get('extended', None) is not None:
                self.raw_cfg.algorithmControl.enableExtended = controls["algorithm_control"]["extended"]
            if controls['algorithm_control'].get('subpixel', None) is not None:
                self.raw_cfg.algorithmControl.enableSubpixel = controls["algorithm_control"]["subpixel"]
            if controls['algorithm_control'].get('lr_check_threshold', None) is not None:
                lrc_threshold = clamp(controls["algorithm_control"]["lr_check_threshold"], *LIMITS['lrc_threshold'])
                self.raw_cfg.algorithmControl.leftRightCheckThreshold = lrc_threshold
            if controls['algorithm_control'].get('subpixel_bits', None) is not None:
                self.raw_cfg.algorithmControl.subpixelFractionalBits = controls["algorithm_control"]["subpixel_bits"]
            if controls['algorithm_control'].get('disparity_shift', None) is not None:
                self.raw_cfg.algorithmControl.disparityShift = controls["algorithm_control"]["disparity_shift"]
            if controls['algorithm_control'].get('invalidate_edge_pixels', None) is not None:
                self.raw_cfg.algorithmControl.numInvalidateEdgePixels = controls["algorithm_control"]["invalidate_edge_pixels"]

        if controls.get('postprocessing', None) is not None:
            if controls['postprocessing'].get('median', None) is not None:
                self.raw_cfg.postProcessing.median = parse_median_filter(controls["postprocessing"]["median"])
            if controls['postprocessing'].get('bilateral_sigma', None) is not None:
                bilateral_sigma = clamp(controls["postprocessing"]["bilateral_sigma"], *LIMITS['bilateral_sigma'])
                self.raw_cfg.postProcessing.bilateralSigmaValue = bilateral_sigma

            if controls['postprocessing'].get('spatial', None) is not None:
                if controls['postprocessing']['spatial'].get('enable', None) is not None:
                    self.raw_cfg.postProcessing.spatialFilter.enable = controls["postprocessing"]["spatial"]["enable"]
                if controls['postprocessing']['spatial'].get('hole_filling', None) is not None:
                    self.raw_cfg.postProcessing.spatialFilter.holeFillingRadius = controls["postprocessing"]["spatial"]["hole_filling"]
                if controls['postprocessing']['spatial'].get('alpha', None) is not None:
                    self.raw_cfg.postProcessing.spatialFilter.alpha = controls["postprocessing"]["spatial"]["alpha"]
                if controls['postprocessing']['spatial'].get('delta', None) is not None:
                    self.raw_cfg.postProcessing.spatialFilter.delta = controls["postprocessing"]["spatial"]["delta"]
                if controls['postprocessing']['spatial'].get('iterations', None) is not None:
                    self.raw_cfg.postProcessing.spatialFilter.numIterations = controls["postprocessing"]["spatial"]["iterations"]

            if controls['postprocessing'].get('temporal', None) is not None:
                if controls['postprocessing']['temporal'].get('enable', None) is not None:
                    self.raw_cfg.postProcessing.temporalFilter.enable = controls["postprocessing"]["temporal"]["enable"]
                if controls['postprocessing']['temporal'].get('persistency_mode', None) is not None:
                    if isinstance(controls["postprocessing"]["temporal"]["persistency_mode"], str):
                        controls["postprocessing"]["temporal"]["persistency_mode"] = getattr(dai.StereoDepthConfig.PostProcessing.TemporalFilter.PersistencyMode, controls["postprocessing"]["temporal"]["persistency_mode"])
                    self.raw_cfg.postProcessing.temporalFilter.persistencyMode = controls["postprocessing"]["temporal"]["persistency_mode"]
                if controls['postprocessing']['temporal'].get('alpha', None) is not None:
                    self.raw_cfg.postProcessing.temporalFilter.alpha = controls["postprocessing"]["temporal"]["alpha"]
                if controls['postprocessing']['temporal'].get('delta', None) is not None:
                    self.raw_cfg.postProcessing.temporalFilter.delta = controls["postprocessing"]["temporal"]["delta"]

            if controls['postprocessing'].get('threshold', None) is not None:
                if controls['postprocessing']['threshold'].get('min_range', None) is not None:
                    min_range = clamp(controls["postprocessing"]["threshold"]["min_range"], *LIMITS['range'])
                    self.raw_cfg.postProcessing.thresholdFilter.minRange = min_range
                if controls['postprocessing']['threshold'].get('max_range', None) is not None:
                    max_range = clamp(controls["postprocessing"]["threshold"]["max_range"], *LIMITS['range'])
                    self.raw_cfg.postProcessing.thresholdFilter.maxRange = max_range

            if controls['postprocessing'].get('brightness', None) is not None:
                if controls['postprocessing']['brightness'].get('min', None) is not None:
                    self.raw_cfg.postProcessing.brightnessFilter.minBrightness = controls["postprocessing"]["brightness"]["min"]
                if controls['postprocessing']['brightness'].get('max', None) is not None:
                    self.raw_cfg.postProcessing.brightnessFilter.maxBrightness = controls["postprocessing"]["brightness"]["max"]

            if controls['postprocessing'].get('speckle', None) is not None:
                if controls['postprocessing']['speckle'].get('enable', None) is not None:
                    self.raw_cfg.postProcessing.speckleFilter.enable = controls["postprocessing"]["speckle"]["enable"]
                if controls['postprocessing']['speckle'].get('range', None) is not None:
                    self.raw_cfg.postProcessing.speckleFilter.speckleRange = controls["postprocessing"]["speckle"]["range"]

            if controls['postprocessing'].get('decimation', None) is not None:
                if controls['postprocessing']['decimation'].get('factor', None) is not None:
                    self.raw_cfg.postProcessing.decimationFilter.decimationFactor = controls["postprocessing"]["decimation"]["factor"]
                if controls['postprocessing']['decimation'].get('mode', None) is not None:
                    if isinstance(controls["postprocessing"]["decimation"]["mode"], str):
                        controls["postprocessing"]["decimation"]["mode"] = getattr(dai.StereoDepthConfig.PostProcessing.DecimationFilter.DecimationMode, controls["postprocessing"]["decimation"]["mode"])
                    self.raw_cfg.postProcessing.decimationFilter.decimationMode = controls["postprocessing"]["decimation"]["mode"]

        if controls.get('census_transform', None) is not None:
            if controls['census_transform'].get('kernel_size', None) is not None:
                if isinstance(controls["census_transform"]["kernel_size"], str):
                    controls["census_transform"]["kernel_size"] = getattr(dai.StereoDepthConfig.CensusTransform.KernelSize, controls["census_transform"]["kernel_size"])
                self.raw_cfg.censusTransform.kernelSize = controls["census_transform"]["kernel_size"]
            if controls['census_transform'].get('kernel_mask', None) is not None:
                self.raw_cfg.censusTransform.kernelMask = controls["census_transform"]["kernel_mask"]
            if controls['census_transform'].get('enable_mean_mode', None) is not None:
                self.raw_cfg.censusTransform.enableMeanMode = controls["census_transform"]["enable_mean_mode"]
            if controls['census_transform'].get('threshold', None) is not None:
                self.raw_cfg.censusTransform.threshold = controls["census_transform"]["threshold"]

        if controls.get('cost_matching', None) is not None:
            if controls['cost_matching'].get('disparity_width', None) is not None:
                if isinstance(controls["cost_matching"]["disparity_width"], str):
                    controls["cost_matching"]["disparity_width"] = getattr(dai.StereoDepthConfig.CostMatching.DisparityWidth, controls["cost_matching"]["disparity_width"])
                self.raw_cfg.costMatching.disparityWidth = controls["cost_matching"]["disparity_width"]
            if controls['cost_matching'].get('enable_companding', None) is not None:
                self.raw_cfg.costMatching.enableCompanding = controls["cost_matching"]["enable_companding"]
            if controls['cost_matching'].get('confidence_threshold', None) is not None:
                conf_threshold = clamp(controls["cost_matching"]["confidence_threshold"], *LIMITS['confidence_threshold'])
                self.raw_cfg.costMatching.confidenceThreshold = conf_threshold
            if controls['cost_matching'].get('linear_equation_parameters', None) is not None:
                if controls['cost_matching']['linear_equation_parameters'].get('alpha', None) is not None:
                    self.raw_cfg.costMatching.linearEquationParameters.alpha = controls["cost_matching"]["linear_equation_parameters"]["alpha"]
                if controls['cost_matching']['linear_equation_parameters'].get('beta', None) is not None:
                    self.raw_cfg.costMatching.linearEquationParameters.beta = controls["cost_matching"]["linear_equation_parameters"]["beta"]
                if controls['cost_matching']['linear_equation_parameters'].get('threshold', None) is not None:
                    self.raw_cfg.costMatching.linearEquationParameters.threshold = controls["cost_matching"]["linear_equation_parameters"]["threshold"]

        if controls.get('cost_aggregation', None) is not None:
            if controls['cost_aggregation'].get('division_factor', None) is not None:
                self.raw_cfg.costAggregation.divisionFactor = controls["cost_aggregation"]["division_factor"]
            if controls['cost_aggregation'].get('horizontal_penalty_cost_p1', None) is not None:
                self.raw_cfg.costAggregation.horizontalPenaltyCostP1 = controls["cost_aggregation"]["horizontal_penalty_cost_p1"]
            if controls['cost_aggregation'].get('horizontal_penalty_cost_p2', None) is not None:
                self.raw_cfg.costAggregation.horizontalPenaltyCostP2 = controls["cost_aggregation"]["horizontal_penalty_cost_p2"]
            if controls['cost_aggregation'].get('vertical_penalty_cost_p1', None) is not None:
                self.raw_cfg.costAggregation.verticalPenaltyCostP1 = controls["cost_aggregation"]["vertical_penalty_cost_p1"]
            if controls['cost_aggregation'].get('vertical_penalty_cost_p2', None) is not None:
                self.raw_cfg.costAggregation.verticalPenaltyCostP2 = controls["cost_aggregation"]["vertical_penalty_cost_p2"]

        ctrl.set(self.raw_cfg)
        self.queue.send(ctrl)



from typing import List

import depthai as dai

from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_imu import XoutIMU


class IMUComponent(Component):
    def __init__(self,
                 device: dai.Device,
                 pipeline: dai.Pipeline):
        self.out = self.Out(self)

        super().__init__()

        self.imu_name: str = device.getConnectedIMU()
        self.node = pipeline.createIMU()
        self.fps = 100
        self.config_imu()  # Default settings, component won't work without them

    def get_imu_name(self) -> str:
        return self.imu_name

    def config_imu(self,
                   sensors: List[dai.IMUSensor] = None,
                   report_rate: int = 100,
                   batch_report_threshold: int = 1,
                   max_batch_reports: int = 10,
                   enable_firmware_update: bool = False
                   ) -> None:
        """
        Configure IMU node.

        Args:
            sensors: List of sensors to enable.
            report_rate: Report rate in Hz.
            batch_report_threshold: Number of reports to batch before sending them to the host.
            max_batch_reports: Maximum number of batched reports to send to the host.
            enable_firmware_update: Enable firmware update if true, disable otherwise.
        """
        sensors = sensors or [dai.IMUSensor.ACCELEROMETER_RAW, dai.IMUSensor.GYROSCOPE_RAW]

        # TODO: check whether we have IMU (and which sensors are supported) once API supports it

        self.node.enableIMUSensor(sensors=sensors, reportRate=report_rate)
        self.node.setBatchReportThreshold(batchReportThreshold=batch_report_threshold)
        self.node.setMaxBatchReports(maxBatchReports=max_batch_reports)
        self.node.enableFirmwareUpdate(enable_firmware_update)

        self.fps = report_rate

    class Out:
        class ImuOut(ComponentOutput):
            def __call__(self, device: dai.Device):
                return XoutIMU(StreamXout(self._comp.node.out, name='imu'), self._comp.fps).set_comp_out(self)

        def __init__(self, imu_component: 'IMUComponent'):
            self.main = self.ImuOut(imu_component)
            self.text = self.main



import os
import os
import time
from threading import Thread
from time import monotonic
from typing import Callable

import depthai as dai

from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.logger import LOGGER
from depthai_sdk.readers.abstract_reader import AbstractReader
from depthai_sdk.utils import *

_fileTypes = ['color', 'left', 'right', 'disparity', 'depth']
_videoExt = ['.mjpeg', '.avi', '.mp4', '.h265', '.h264', '.webm']
_imageExt = ['.bmp', '.dib', '.jpeg', '.jpg', '.jpe', '.jp2', '.png', '.webp', '.pbm', '.pgm', '.ppm', '.pxm',
             '.pnm', '.pfm', '.sr', '.ras', '.tiff', '.tif', '.exr', '.hdr', '.pic']


def _run(delay: float, sendFrames: Callable):
    while True:
        if not sendFrames():
            break
        time.sleep(delay)
    LOGGER.info('Replay `run` thread stopped')


class ReplayStream:
    @property
    def shape(self) -> Tuple[int, int]:
        return self.resize if self.resize else self._shape

    def __init__(self):
        self.node: dai.node.XLinkIn = None
        self.queue: dai.DataInputQueue = None
        self.disabled = False
        self.stream_name = ''
        self.camera_socket: dai.CameraBoardSocket = None  # Forced socket

        self.resize: Tuple[int, int] = None
        self.resize_mode: ResizeMode = None
        self._shape: Tuple[int, int] = None
        self.callbacks: List[Callable] = []

        self.frame: np.ndarray  # Last read frame from Reader (ndarray)
        self.imgFrame: dai.ImgFrame  # Last read ImgFrame from Reader (dai.ImgFrame)
        self.size_bytes: int  # bytes

    def get_socket(self) -> dai.CameraBoardSocket:
        if self.camera_socket is not None:
            return self.camera_socket
        if 'left' in self.stream_name.lower():
            return dai.CameraBoardSocket.LEFT
        elif 'right' in self.stream_name.lower():
            return dai.CameraBoardSocket.RIGHT
        else:
            return dai.CameraBoardSocket.CAM_A
        # raise Exception("Please specify replay stream CameraBoardSocket via replay.specify_socket()")


class Replay:
    def __init__(self, path: Union[Path, str]):
        """
        Helper file to replay recorded depthai stream. It reads from recorded files (mjpeg/avi/mp4/h265/h264/bag)
        and sends frames back to OAK camera to replay the scene, including depth reconstruction from 2 synced mono
        streams.

        Args:
            path (str): Path to the recording folder.
        """
        self.path = self._get_path(path)

        self.disabledStreams: List[str] = []
        self.streams: Dict[str, ReplayStream] = {}

        self._inputQueues = dict()  # dai.InputQueue dictionary for each stream
        self._seqNum = 0  # Frame sequence number, added to each imgFrame
        self._now: monotonic = None
        self._colorSize = None
        self._keepAR = True  # By default, crop image as needed to keep the aspect ratio
        self._pause = False
        self._calibData = None

        self.fps: float = 30.0
        self.thread: Optional[Thread] = None
        self._stop: bool = False  # Stop the thread that's sending frames to the OAK camera

        self.xins: List[str] = []  # Name of XLinkIn streams

        self.reader: Optional[AbstractReader] = None
        self.frames: Dict[str, np.ndarray] = dict()  # Frames read from Readers
        self.imgFrames: Dict[str, dai.ImgFrame] = dict()  # Last frame sent to the device

        def cntFilesExt(path: Path, ext: Union[str, List[str]]) -> int:
            def fileWithExt(file: str) -> bool:
                if isinstance(ext, List):
                    return os.path.splitext(file)[1] in ext
                elif isinstance(ext, str):
                    return file.endswith(ext)
                else:
                    raise ValueError('ext should be either str or List[str]!')

            return [fileWithExt(f) for f in os.listdir(str(path))].count(True)

        if self.path.is_dir():  # Provided path is a folder
            if 0 < cntFilesExt(self.path, _imageExt):
                from .readers.image_reader import ImageReader
                self.reader = ImageReader(self.path)
            elif 0 < cntFilesExt(self.path, _videoExt):
                from .readers.videocap_reader import VideoCapReader
                self.reader = VideoCapReader(self.path)
            elif cntFilesExt(self.path, '.bag') == 1:
                from .readers.rosbag_reader import RosbagReader
                self.reader = RosbagReader(self.path)
            elif cntFilesExt(self.path, '.mcap') == 1:
                from .readers.mcap_reader import McapReader
                self.reader = McapReader(self.path)
            elif cntFilesExt(self.path, '.db3') == 1:
                from .readers.db3_reader import Db3Reader
                self.reader = Db3Reader(self.path)
            else:
                raise RuntimeError("Path invalid - no recordings found.")

            # Read calibration file
            calib_file = self.path / 'calib.json'
            if calib_file.exists():
                self._calibData = dai.CalibrationHandler(str(calib_file))

        else:  # Provided path is a file
            if self.path.suffix in _videoExt:
                from .readers.videocap_reader import VideoCapReader
                self.reader = VideoCapReader(self.path)
            elif self.path.suffix in _imageExt:
                from .readers.image_reader import ImageReader
                self.reader = ImageReader(self.path)
            else:
                raise NotImplementedError('Please select folder')

        # Read all available streams
        for stream_name in self.reader.getStreams():
            stream = ReplayStream()
            stream._shape = self.reader.getShape(stream_name)
            stream.size_bytes = self.reader.get_message_size(stream_name)
            stream.camera_socket = self.reader.get_socket(stream_name)
            self.streams[stream_name.lower()] = stream

    def _get_path(self, path: str) -> Path:
        """
        Either use local depthai-recording, YT link, mp4 url
        @param path: depthai-recording path.
        @return: Replay module
        """
        if isinstance(path, Path):
            return path.resolve()
        if isUrl(path):
            if isYoutubeLink(path):
                # Overwrite source - so Replay class can use it
                return downloadYTVideo(path)
            else:
                return downloadContent(path)

        if Path(path).resolve().exists():
            return Path(path).resolve()

        recording_name: str = path
        # Check if we have it stored locally
        path: Path = getLocalRecording(recording_name)
        if path is not None:
            return path

        # Try to download from the server
        dic = getAvailableRecordings()
        if recording_name in dic:
            arr = dic[recording_name]
            LOGGER.info("Downloading depthai recording '{}' from Luxonis' servers, in total {:.2f} MB"
                         .format(recording_name, arr[1] / 1e6))
            path = downloadRecording(recording_name, arr[0])
            return path
        else:
            raise ValueError(f"DepthAI recording '{recording_name}' was not found on the server!")

    def toggle_pause(self):
        """
        Toggle pausing of sending frames to the OAK camera.
        """
        self._pause = not self._pause

    def set_fps(self, fps: float):
        """
        Sets frequency at which Replay module will send frames to the camera. Default 30FPS.
        """
        if type(self.reader).__name__ == 'ImageReader':
            self.reader.set_cycle_fps(fps)
        else:
            self.fps = fps

    def set_loop(self, flag: bool):
        """
        Sets whether to loop the replay.

        Args:
            flag (bool): Whether to loop the replay.
        """
        from .readers.videocap_reader import VideoCapReader
        if isinstance(self.reader, VideoCapReader):
            self.reader.set_loop(flag)
        else:
            raise RuntimeError('Looping is only supported for video files.')

    def get_fps(self) -> float:
        return self.fps

    def _add_callback(self, stream_name: str, callback: Callable):
        self.streams[stream_name.lower()].callbacks.append(callback)

    def resize(self, stream_name: str, size: Tuple[int, int], mode: ResizeMode = ResizeMode.STRETCH):
        """
        Resize color frames prior to sending them to the device.

        Args:
            stream_name (str): Name of the stream we want to resize
            size (Tuple(width, heigth)): Size of color frames that are sent to the camera
            mode (ResizeMode): How to actually resize the stream
        """
        self.streams[stream_name.lower()].resize = size
        self.streams[stream_name.lower()].resize_mode = mode

    def keepAspectRatio(self, keepAspectRatio: bool):
        raise Exception('keepAspectRatio() has been deprecated, use resize(mode=ResizeMode) to set whether to keep AR!')

    def disableStream(self, stream_name: str, disableReading: bool = False):
        """
        Disable sending a recorded stream to the device.

        Args:
            streamName(str): Name of the stream to disable (eg. 'left', 'color', 'depth', etc.)
            disableReading (bool, Optional): Also disable reading frames from the file
        """
        if disableReading:
            self.reader.disableStream(stream_name)

        if stream_name.lower() not in self.streams:
            LOGGER.info(f"There's no stream '{stream_name}' available!")
            return

        self.streams[stream_name.lower()].disabled = True

    def specify_socket(self, stream_name: str, socket: dai.CameraBoardSocket):
        if stream_name.lower() not in self.streams:
            LOGGER.info(f"There's no stream '{stream_name}' available!")
            return
        self.streams[stream_name.lower()].camera_socket = socket

    def initPipeline(self, pipeline: dai.Pipeline = None):
        """
        Prepares the pipeline for replaying. It creates XLinkIn nodes and sets up StereoDepth node.
        Returns: dai.Pipeline
        """
        if pipeline is None:  # Create pipeline if not passed
            pipeline = dai.Pipeline()

        if self._calibData is not None:
            pipeline.setCalibrationData(self._calibData)

        def createXIn(p: dai.Pipeline, xlink_stream_name: str, size: int):
            xin = p.create(dai.node.XLinkIn)
            xin.setMaxDataSize(size)
            xin.setStreamName(xlink_stream_name)
            return xin

        for name, stream in self.streams.items():
            if stream.disabled: continue

            stream.stream_name = name + '_in'
            stream.node = createXIn(pipeline, stream.stream_name, stream.size_bytes)

        return pipeline

    def initStereoDepth(self,
                        stereo: dai.node.StereoDepth,
                        left_name: str = 'left',
                        right_name: str = 'right',
                        align_to: str = ''):
        if left_name.lower() not in self.streams or right_name.lower() not in self.streams:
            raise Exception("Tried to init StereoDepth, but left/right streams aren't available!")

        left = self.streams[left_name.lower()]
        right = self.streams[right_name.lower()]

        stereo.setInputResolution(left.shape)

        if not left.camera_socket:
            left.camera_socket = dai.CameraBoardSocket.LEFT
        if not right.camera_socket:
            right.camera_socket = dai.CameraBoardSocket.RIGHT

        if align_to:  # Enable RGB-depth alignment
            stereo.setDepthAlign(dai.CameraBoardSocket.RGB)
            stereo.setOutputSize(*self.streams[align_to.lower()].shape)

        left.node.out.link(stereo.left)
        right.node.out.link(stereo.right)

    def start(self):
        """
        Start sending frames to the OAK device on a new thread
        """
        self.thread = Thread(target=_run, args=(1.0 / self.fps, self.sendFrames,))
        self.thread.start()

    def sendFrames(self) -> bool:
        """
        Reads and sends recorded frames from all enabled streams to the OAK camera.

        Returns:
            bool: True if successful, otherwise False.
        """
        if not self._pause:  # If replaying is paused, don't read new frames
            if not self._readFrames():
                self._stop = True
                return False  # End of the recording

        self._now = monotonic()
        for stream_name, stream in self.streams.items():
            stream.imgFrame = self._createImgFrame(stream)
            # Save the imgFrame
            for cb in stream.callbacks:  # callback
                cb(stream_name.lower(), stream.imgFrame)

            # Don't send these frames to the OAK camera
            if stream.disabled:
                continue

            # Send an imgFrame to the OAK camera
            stream.queue.send(stream.imgFrame)

        self._seqNum += 1
        return True

    def createQueues(self, device: dai.Device):
        """
        Creates input queue for each enabled stream

        Args:
            device (dai.Device): Device to which we will stream frames
        """

        for _, stream in self.streams.items():
            if stream.stream_name:
                stream.queue = device.getInputQueue(stream.stream_name)

    def getStreams(self) -> List[str]:
        return [name.lower() for name, _ in self.streams.items()]

    def _resize_frame(self, frame: np.ndarray, size: Tuple[int, int], mode: ResizeMode) -> np.ndarray:
        if mode == ResizeMode.STRETCH:
            # No need to keep aspect ratio, image will be squished
            return cv2.resize(frame, size)
        elif mode == ResizeMode.CROP:
            cropped = cropToAspectRatio(frame, size)
            return cv2.resize(cropped, size)
        elif mode == ResizeMode.FULL_CROP:
            w = frame.shape[1]
            start_w = int((w - size[0]) / 2)
            h = frame.shape[0]
            start_h = int((h - size[1]) / 2)
            return frame[start_h:h - start_h, start_w:w - start_w]

    def _createNewFrame(self, cvFrame) -> dai.ImgFrame:
        imgFrame = dai.ImgFrame()
        imgFrame.setData(cvFrame)
        imgFrame.setTimestamp(self._now)
        imgFrame.setSequenceNum(self._seqNum)
        shape = cvFrame.shape[::-1]
        imgFrame.setWidth(shape[0])
        imgFrame.setHeight(shape[1])
        return imgFrame

    def _createImgFrame(self, stream: ReplayStream) -> dai.ImgFrame:
        cvFrame: np.ndarray = stream.frame
        if stream.resize:
            cvFrame = self._resize_frame(cvFrame, stream.resize, stream.resize_mode)

        if cvFrame.shape[-1] == 3:  # 3 channels = RGB
            # Resize/crop color frame as specified by the user
            # cv2 reads frames in interleaved format, and most networks expect planar by default
            cvFrame = toPlanar(cvFrame)
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.BGR888p)
            imgFrame.setInstanceNum(int(stream.get_socket()))
        elif cvFrame.dtype == np.uint8:
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.RAW8)
            imgFrame.setInstanceNum(int(stream.get_socket()))
        elif cvFrame.dtype == np.uint16:
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.RAW16)
        else:
            raise Exception('Unknown frame types')

        return imgFrame

    def _readFrames(self) -> bool:
        """
        Reads frames from all Readers.

        Returns:
            bool: True if successful, otherwise False.
        """
        frames = self.reader.read()
        if not frames:
            return False  # No more frames!

        for name, frame in frames.items():
            self.streams[name.lower()].frame = frame

        # Compress 3-plane frame to a single plane
        # for name, frame in self.frames.items():
        #     if name in ["left", "right", "disparity"] and len(frame.shape) == 3:
        #         self.frames[name] = frame[:, :, 0]  # All 3 planes are the same
        return True

    def getShape(self, name: str) -> Tuple[int, int]:
        """
        Get shape of a stream
        """
        return self.streams[name.lower()].shape

    def close(self):
        """
        Closes all video readers.
        """
        self._stop = True
        self.reader.close()
        if self.thread:
            self.thread.join()



from typing import List, Optional, Union, Tuple

import depthai as dai

from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.parser import parse_camera_socket
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout, XoutBase
from depthai_sdk.oak_outputs.xout.xout_depth import XoutDisparityDepth
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.components.tof_control import ToFControl


class ToFComponent(Component):

    _align: Optional[dai.node.ImageAlign] = None

    def __init__(
        self,
        device: dai.Device,
        pipeline: dai.Pipeline,
        source: Union[str, dai.CameraBoardSocket, None] = None,
        align_to: Optional[CameraComponent] = None,
        fps: Optional[int] = None,
    ):
        super().__init__()
        self.out = self.Out(self)
        self._pipeline = pipeline

        if source is None:
            source = self._find_tof(device)
        elif isinstance(source, str):
            source = parse_camera_socket(source)
        elif isinstance(source, dai.CameraBoardSocket):
            pass  # This is what we want
        else:
            raise ValueError("source must be either None, str, or CameraBoardSocket!")

        self.control = ToFControl(device)
        self.camera_node = pipeline.create(dai.node.ColorCamera)
        self.camera_node.setBoardSocket(source)
        self.camera_socket = source

        if fps is not None:
            self.camera_node.setFps(fps)

        self.node: dai.node.ToF = pipeline.create(dai.node.ToF)
        self._control_in = pipeline.create(dai.node.XLinkIn)
        self.camera_node.raw.link(self.node.input)
        self._control_in.setStreamName(f"{self.node.id}_inputControl")
        self._control_in.out.link(self.node.inputConfig)

        self._align_to_output: dai.Node.Output = None

        if align_to is not None:
            self.set_align_to(align_to)

    def _find_tof(self, device: dai.Device) -> dai.CameraBoardSocket:
        # Use the first ToF sensor, usually, there will only be one
        features = device.getConnectedCameraFeatures()
        for cam_sensor in features:
            if dai.CameraSensorType.TOF in cam_sensor.supportedTypes:
                return cam_sensor.socket
        raise RuntimeError(
            f"No ToF sensor found on this camera! {device.getConnectedCameraFeatures()}"
        )

    def set_align_to(self, align_to: CameraComponent, output_size: Optional[Tuple] = None) -> None:
        if self._align is None:
            self._align = self._pipeline.create(dai.node.ImageAlign)
            self.node.depth.link(self._align.input)
            if align_to.is_mono():
                self._align_to_output = align_to.node.out
            else:
                self._align_to_output = align_to.node.isp

            self._align_to_output.link(self._align.inputAlignTo)

            if output_size is not None:
                self._align.setOutputSize(*output_size)


    def configure_tof(self,
                        phaseShuffleTemporalFilter: bool = None,
                        phaseUnwrappingLevel: int = None,
                        phaseUnwrapErrorThreshold: int = None,
                      ) -> None:
        tofConfig = self.node.initialConfig.get()
        if phaseShuffleTemporalFilter is not None:
            tofConfig.enablePhaseShuffleTemporalFilter = phaseShuffleTemporalFilter
        if phaseUnwrappingLevel is not None:
            tofConfig.phaseUnwrappingLevel = phaseUnwrappingLevel
        if phaseUnwrapErrorThreshold is not None:
            tofConfig.phaseUnwrapErrorThreshold = phaseUnwrapErrorThreshold
        self.node.initialConfig.set(tofConfig)

    def on_pipeline_started(self, device: dai.Device) -> None:
        self.control.set_input_queue(
            device.getInputQueue(self._control_in.getStreamName())
        )

    class Out:

        class DepthOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutDisparityDepth(
                    device=device,
                    frames=StreamXout(
                        (
                            self._comp.node.depth
                            if self._comp._align is None
                            else self._comp._align.outputAligned
                        ),
                        "tof_depth",
                    ),
                    aligned_frame=StreamXout(self._comp._align_to_output, "aligned_stream") if self._comp._align else None,
                    dispScaleFactor=9500,
                    ir_settings={"auto_mode": False},
                ).set_comp_out(self)

        class AmplitudeOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(
                    frames=StreamXout(self._comp.node.amplitude, "tof_amplitude")
                ).set_comp_out(self)

        class ErrorOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(
                    frames=StreamXout(self._comp.node.error, "tof_error")
                ).set_comp_out(self)

        def __init__(self, tof_component: "ToFComponent"):
            self.depth = self.DepthOut(tof_component)
            self.amplitude = self.AmplitudeOut(tof_component)
            self.error = self.ErrorOut(tof_component)
            self.main = self.depth



from typing import Optional, Union, Any

import depthai as dai

from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.components.tof_component import ToFComponent
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout
from depthai_sdk.oak_outputs.xout.xout_pointcloud import XoutPointcloud
from depthai_sdk.replay import Replay


class PointcloudComponent(Component):

    def __init__(self,
                 device: dai.Device,
                 pipeline: dai.Pipeline,
                 depth_input: Union[None, StereoComponent, ToFComponent, dai.node.StereoDepth, dai.Node.Output] = None,
                 colorize: Optional[CameraComponent] = None,
                 replay: Optional[Replay] = None,
                 args: Any = None):
        """
        Args:
            pipeline (dai.Pipeline): DepthAI pipeline
            replay (Replay object, optional): Replay
            args (Any, optional): Use user defined arguments when constructing the pipeline
        """
        super().__init__()
        self.out = self.Out(self)

        self.depth: dai.Node.Output  # depth output

        self.colorize_comp: Optional[CameraComponent] = colorize

        self._replay: Optional[Replay] = replay

        # Depth aspect
        if depth_input is None:
            depth_input = StereoComponent(device, pipeline, replay=replay, args=args)
            depth_input.config_stereo(lr_check=True, subpixel=True, subpixel_bits=3, confidence=230)
            depth_input.node.initialConfig.setNumInvalidateEdgePixels(20)

            config = depth_input.node.initialConfig.get()
            config.postProcessing.speckleFilter.enable = True
            config.postProcessing.speckleFilter.speckleRange = 50
            config.postProcessing.temporalFilter.enable = True
            config.postProcessing.spatialFilter.enable = True
            config.postProcessing.spatialFilter.holeFillingRadius = 2
            config.postProcessing.spatialFilter.numIterations = 1
            config.postProcessing.thresholdFilter.minRange = 400  # 40cm
            config.postProcessing.thresholdFilter.maxRange = 20000  # 20m
            config.postProcessing.decimationFilter.decimationFactor = 2
            config.postProcessing.decimationFilter.decimationMode = dai.RawStereoDepthConfig.PostProcessing.DecimationFilter.DecimationMode.NON_ZERO_MEDIAN
            depth_input.node.initialConfig.set(config)

            if self.colorize_comp is not None:
                # Align to colorize node
                depth_input.config_stereo(align=self.colorize_comp)

        if isinstance(depth_input, StereoComponent):
            depth_input = depth_input.node
        elif isinstance(depth_input, ToFComponent):
            if depth_input._align is not None:
                self.depth = depth_input._align.outputAligned
            else:
                self.depth = depth_input.node.depth

        if isinstance(depth_input, dai.node.StereoDepth):
            self.depth = depth_input.depth
        elif isinstance(depth_input, dai.Node.Output):
            self.depth = depth_input

    def config_postprocessing(self) -> None:
        """
        Configures postprocessing options.
        """
        raise NotImplementedError("config_postprocessing() not yet implemented")

    class Out:
        class PointcloudOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                colorize = None
                if self._comp.colorize_comp is not None:
                    colorize = StreamXout(self._comp.colorize_comp.stream, name="Color")
                return XoutPointcloud(device,
                                      StreamXout(self._comp.depth),
                                      color_frames=colorize).set_comp_out(self)

        def __init__(self, component: 'PointcloudComponent'):
            self.pointcloud = self.PointcloudOut(component)
            self.main = self.pointcloud



from enum import Enum
from typing import List

import depthai as dai
import numpy as np


class ImuSyncMethod(Enum):
    LINEAR_INTERPOLATE_ACCEL = 'LINEAR_INTERPOLATE_ACCEL'
    LINEAR_INTERPOLATE_GYRO = 'LINEAR_INTERPOLATE_GYRO'
    COPY = 'COPY'


class ImuInterpolation:
    def __init__(self):
        self.imu_packets: List[dai.IMUPacket] = []

    def Imu(self, msg, imu_packet: dai.IMUPacket,
            sync_mode: ImuSyncMethod = ImuSyncMethod.LINEAR_INTERPOLATE_ACCEL,
            linear_accel_cov: float = 0., angular_velocity_cov: float = 0.):
        # TODO: rather pass classes (Imu, Quaternion, Vector3) and create object here?
        # When passing ros_imu_msg make sure all attributes are already defined!

        self.imu_packets.append(imu_packet)
        if 20 < len(self.imu_packets):
            self.imu_packets.pop(0)

        if sync_mode != ImuSyncMethod.COPY:
            interp_imu_packets = self.fillImuData_LinearInterpolation(sync_mode)
            if 0 < len(interp_imu_packets):
                imu_packet = interp_imu_packets[-1]

        if imu_packet.acceleroMeter is not None:
            msg.linear_acceleration.x = imu_packet.acceleroMeter.x
            msg.linear_acceleration.y = imu_packet.acceleroMeter.y
            msg.linear_acceleration.z = imu_packet.acceleroMeter.z

        if imu_packet.gyroscope is not None:
            msg.angular_velocity.x = imu_packet.gyroscope.x
            msg.angular_velocity.y = imu_packet.gyroscope.y
            msg.angular_velocity.z = imu_packet.gyroscope.z

        msg.linear_acceleration_covariance = np.array([linear_accel_cov, 0.0, 0.0, 0.0, linear_accel_cov, 0.0, 0.0, 0.0,
                                                       linear_accel_cov])
        msg.angular_velocity_covariance = np.array([angular_velocity_cov, 0.0, 0.0, 0.0, angular_velocity_cov, 0.0,
                                                    0.0, 0.0, angular_velocity_cov])

        msg.orientation.x = 0.0
        msg.orientation.y = 0.0
        msg.orientation.z = 0.0
        msg.orientation.w = 0.0

        msg.orientation_covariance = np.array([-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

    def _lerp(self, a, b, t):
        return a * (1.0 - t) + b * t

    def _lerpImu(self, a, b, t):
        res = a.__class__()
        res.x = self._lerp(a.x, b.x, t)
        res.y = self._lerp(a.y, b.y, t)
        res.z = self._lerp(a.z, b.z, t)
        return res

    def fillImuData_LinearInterpolation(self, sync_mode: ImuSyncMethod):
        accel_hist = []
        gyro_hist = []
        interp_imu_packets = []

        for i in range(len(self.imu_packets)):
            if len(accel_hist) == 0:
                accel_hist.append(self.imu_packets[i].acceleroMeter)
            elif accel_hist[-1].sequence != self.imu_packets[i].acceleroMeter.sequence:
                accel_hist.append(self.imu_packets[i].acceleroMeter)

            if len(gyro_hist) == 0:
                gyro_hist.append(self.imu_packets[i].gyroscope)
            elif gyro_hist[-1].sequence != self.imu_packets[i].gyroscope.sequence:
                gyro_hist.append(self.imu_packets[i].gyroscope)

            if sync_mode.value == ImuSyncMethod.LINEAR_INTERPOLATE_ACCEL:
                if len(accel_hist) < 3:
                    continue
                else:
                    accel0 = dai.IMUReportAccelerometer()
                    accel0.sequence = -1

                    while len(accel_hist) > 0:
                        if accel0.sequence == -1:
                            accel0 = accel_hist.pop(0)
                        else:
                            accel1 = accel_hist.pop(0)
                            dt = (accel1.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000

                            while len(gyro_hist) > 0:
                                curr_gyro = gyro_hist[0]

                                if curr_gyro.timestamp.get() > accel0.timestamp.get() and curr_gyro.timestamp.get() <= accel1.timestamp.get():
                                    diff = (curr_gyro.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000
                                    alpha = diff / dt
                                    interp_accel = self._lerpImu(accel0, accel1, alpha)
                                    imu_packet = dai.IMUPacket()
                                    imu_packet.acceleroMeter = interp_accel
                                    imu_packet.gyroscope = curr_gyro
                                    interp_imu_packets.append(imu_packet)
                                    gyro_hist.pop(0)

                                elif curr_gyro.timestamp.get() > accel1.timestamp.get():
                                    accel0 = accel1
                                    if len(accel_hist) > 0:
                                        accel1 = accel_hist.pop(0)
                                        dt = (accel1.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000
                                    else:
                                        break
                                else:
                                    gyro_hist.pop(0)

                            accel0 = accel1

                    accel_hist.append(accel0)

            elif sync_mode == ImuSyncMethod.LINEAR_INTERPOLATE_GYRO:
                if len(gyro_hist) < 3:
                    continue
                else:
                    gyro0 = dai.IMUReportGyroscope()
                    gyro0.sequence = -1

                    while len(gyro_hist) > 0:
                        if gyro0.sequence == -1:
                            gyro0 = gyro_hist.pop(0)
                        else:
                            gyro1 = gyro_hist.pop(0)
                            dt = (gyro1.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000

                            while len(accel_hist) > 0:
                                curr_accel = accel_hist[0]

                                if curr_accel.timestamp.get() > gyro0.timestamp.get() and curr_accel.timestamp.get() <= gyro1.timestamp.get():
                                    diff = (curr_accel.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000
                                    alpha = diff / dt
                                    interp_gyro = self._lerpImu(gyro0, gyro1, alpha)
                                    imu_packet = dai.IMUPacket()
                                    imu_packet.acceleroMeter = curr_accel
                                    imu_packet.gyroscope = interp_gyro
                                    interp_imu_packets.append(imu_packet)
                                    accel_hist.pop(0)

                                elif curr_accel.timestamp.get() > gyro1.timestamp.get():
                                    gyro0 = gyro1

                                    if len(gyro_hist) > 0:
                                        gyro1 = gyro_hist.pop(0)
                                        dt = (gyro1.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000
                                    else:
                                        break

                                else:
                                    accel_hist.pop(0)

                            gyro0 = gyro1

                    gyro_hist.append(gyro0)

        return interp_imu_packets
